''' Tank exit code check plugin '''
from ...common.interfaces import AbstractPlugin


class Plugin(AbstractPlugin):
    SECTION = 'rcassert'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.ok_codes = []
        self.fail_code = 10

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return ["pass", "fail_code"]

    def configure(self):
        codes = self.get_option("pass", '').split(' ')
        for code in codes:
            if code:
                self.ok_codes.append(int(code))
        self.fail_code = int(self.get_option("fail_code", self.fail_code))

    def post_process(self, retcode):
        if not self.ok_codes:
            return retcode

        for code in self.ok_codes:
            self.log.debug("Comparing %s with %s codes", code, retcode)
            if code == int(retcode):
                self.log.info("Exit code %s was changed to 0 by RCAssert plugin", code)
                return 0

        self.log.info("Changing exit code to %s because RCAssert pass list was unsatisfied", self.fail_code)
        return self.fail_code



import datetime

from ...common.interfaces import AbstractInfoWidget


class BfgInfoWidget(AbstractInfoWidget):
    ''' Console widget '''

    def __init__(self):
        AbstractInfoWidget.__init__(self)
        self.active_threads = 0
        self.instances = 0
        self.planned = 0
        self.RPS = 0
        self.selfload = 0
        self.time_lag = 0
        self.planned_rps_duration = 0

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stat):
        self.instances = stat["metrics"]["instances"]

        self.RPS = data["overall"]["interval_real"]["len"]
        self.selfload = 0  # TODO
        self.time_lag = 0  # TODO

    def render(self, screen):
        res = ''

        res += "Active instances: "
        res += str(self.instances)

        res += "\nPlanned requests: %s for %s\nActual responses: " % (
            self.planned, datetime.timedelta(seconds=self.planned_rps_duration))
        if self.planned != self.RPS:
            res += screen.markup.YELLOW + str(self.RPS) + screen.markup.RESET
        else:
            res += str(self.RPS)

        res += "\n        Accuracy: "
        if self.selfload < 80:
            res += screen.markup.RED + \
                ('%.2f' % self.selfload) + screen.markup.RESET
        elif self.selfload < 95:
            res += screen.markup.YELLOW + \
                ('%.2f' % self.selfload) + screen.markup.RESET
        else:
            res += ('%.2f' % self.selfload)

        res += "%\n        Time lag: "
        res += str(datetime.timedelta(seconds=self.time_lag))

        return res



import logging
import sys
import multiprocessing
import threading
from optparse import OptionParser

from yandextank.contrib.netort.netort.resource import manager as resource_manager
from yandextank.core.tankworker import TankWorker
from yandextank.core.tankcore import LockError
from yandextank.validator.validator import ValidationError
from yandextank.version import VERSION


def main():
    parser = OptionParser()
    parser.add_option(
        '-c',
        '--config',
        action='append',
        help="Path to INI file containing run options, multiple options accepted",
        default=[]
    )
    parser.add_option(
        '-f',
        '--fail-lock',
        action='store_true',
        dest='lock_fail',
        help="Don't wait for lock to release, fail test instead")
    parser.add_option(
        '-i',
        '--ignore-lock',
        action='store_true',
        dest='ignore_lock',
        help="Ignore lock files from concurrent instances, has precedence before --lock-fail"
    )
    parser.add_option(
        '-k',
        '--lock-dir',
        action='store',
        dest='lock_dir',
        type="string",
        help="Directory for lock file")
    parser.add_option(
        '-l',
        '--log',
        action='store',
        default="tank.log",
        help="Tank log file location")
    parser.add_option(
        '--error_log',
        action='store',
        dest='error_log',
        default="tank_errors.log",
        help="Tank errors log file location")
    parser.add_option(
        '-m',
        '--manual-start',
        action='store_true',
        dest='manual_start',
        help="Wait for Enter key to start the test")
    parser.add_option(
        '-n',
        '--no-rc',
        action='store_true',
        dest='no_rc',
        help="Don't load config files from /etc/yandex-tank and ~/.yandex-tank")
    parser.add_option(
        '-o',
        '--option',
        action='append',
        help="Set config option, multiple options accepted, example: -o 'shellexec.start=pwd'"
    )
    parser.add_option(
        '-q',
        '--quiet',
        action='store_true',
        help="Less console output, only errors and warnings")
    parser.add_option(
        '-s',
        '--scheduled-start',
        action='store',
        dest='scheduled_start',
        help="Start test at specified time, format 'YYYY-MM-DD hh:mm:ss', date part is optional"
    )
    parser.add_option(
        '-v',
        '--verbose',
        action='store_true',
        help="More console output, +debug messages")
    parser.add_option(
        '-p',
        '--patch-cfg',
        action='append',
        help='Patch config with yaml snippet (similar to -o, but has full compatibility to\
        and the exact scheme of yaml format config)',
        dest='patches'
    )
    parser.add_option(
        '--version',
        action='store_true',
        dest='version'
    )
    # FIXME: restore completion helper
    # completion_helper = CompletionHelperOptionParser()
    # completion_helper.handle_request(parser)

    options, ammofiles = parser.parse_args()
    if options.version:
        print('YandexTank/{}'.format(VERSION))
        return

    ammofile = ammofiles[0] if len(ammofiles) > 0 else None

    handlers = init_logging(options.error_log, options.verbose, options.quiet)

    cli_kwargs = {'core': {'lock_dir': options.lock_dir}} if options.lock_dir else {}
    if options.ignore_lock:
        cli_kwargs.setdefault('core', {})['ignore_lock'] = options.ignore_lock

    if ammofile:
        logging.debug("Ammofile: %s", ammofile)
        cli_kwargs['phantom'] = {
            'use_caching': False,
            'ammofile': ammofile
        }

    run_shooting_event = multiprocessing.Event() if options.manual_start else None

    try:
        worker = TankWorker([resource_manager.resource_filename(cfg) for cfg in options.config],
                            options.option,
                            options.patches,
                            [cli_kwargs],
                            options.no_rc,
                            ammo_file=ammofile if ammofile else None,
                            log_handlers=handlers,
                            debug=options.verbose,
                            run_shooting_event=run_shooting_event,
                            )
    except (ValidationError, LockError) as e:
        logging.error(f'Config validation error:\n{e}')
        return
    worker.start()
    try:
        if run_shooting_event is not None:
            # Parallel process(TankWorker) is logging.
            # So, prompt should be repeated.
            # Otherwise prompt will be lost among parallel logging.
            def wait_blocking():
                input()
                run_shooting_event.set()
            threading.Thread(target=wait_blocking).start()
            while not run_shooting_event.wait(3):
                logging.warning('Press Enter to start shooting')
        while True:
            worker.join(timeout=2)
            if not worker.is_alive():
                break
    except KeyboardInterrupt:
        worker.stop()
        worker.join()
    sys.exit(worker.retcode)


def init_logging(events_log_fname, verbose, quiet):
    """ Set up logging, as it is very important for console tool """
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)

    # create file handler which logs error messages

    # create console handler with a higher log level
    console_handler = logging.StreamHandler(sys.stdout)
    stderr_hdl = logging.StreamHandler(sys.stderr)

    fmt_verbose = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)d\t%(message)s"
    )
    fmt_regular = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(message)s", "%H:%M:%S")

    if verbose:
        console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(fmt_verbose)
        stderr_hdl.setFormatter(fmt_verbose)
    elif quiet:
        console_handler.setLevel(logging.WARNING)
        console_handler.setFormatter(fmt_regular)
        stderr_hdl.setFormatter(fmt_regular)
    else:
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(fmt_regular)
        stderr_hdl.setFormatter(fmt_regular)

    f_err = SingleLevelFilter(logging.ERROR, True)
    f_warn = SingleLevelFilter(logging.WARNING, True)
    f_crit = SingleLevelFilter(logging.CRITICAL, True)
    console_handler.addFilter(f_err)
    console_handler.addFilter(f_warn)
    console_handler.addFilter(f_crit)
    logger.addHandler(console_handler)

    f_info = SingleLevelFilter(logging.INFO, True)
    f_debug = SingleLevelFilter(logging.DEBUG, True)
    stderr_hdl.addFilter(f_info)
    stderr_hdl.addFilter(f_debug)
    logger.addHandler(stderr_hdl)

    if events_log_fname:
        err_file_handler = logging.FileHandler(events_log_fname)
        err_file_handler.setLevel(logging.WARNING)
        err_file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s\t%(message)s"
            ))
        return [err_file_handler, console_handler, stderr_hdl]
    else:
        return [console_handler, stderr_hdl]
        # logger.addHandler(err_file_handler)


class SingleLevelFilter(logging.Filter):
    """Exclude or approve one msg type at a time.    """

    def __init__(self, passlevel, reject):
        logging.Filter.__init__(self)
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return record.levelno != self.passlevel
        else:
            return record.levelno == self.passlevel


if __name__ == '__main__':
    main()



import re
from collections import defaultdict
import logging

LOGGER = logging.getLogger(__file__)


class LogFormatError(Exception):
    ...


LINE_FORMAT = r'(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d.\d\d\d [+-]\d\d\d\d) \[([^\]]+)\] \[([^\]]*)\] (.*)'


class LogLine:
    def __init__(self, line: str):
        mach = re.match(LINE_FORMAT, line)
        if mach is None:
            raise LogFormatError(f"Phantom log line doesn't match format: {LINE_FORMAT}")
        self.time_stamp, self.level, self.name, message = mach.groups()
        self.message = message.strip()


class LogAnalyzer:
    def __init__(self, path):
        self.path = path

    def get_most_recent_errors(self, limit=10):
        counter = defaultdict(int)
        with open(self.path) as f:
            for line in f:
                try:
                    parsed = LogLine(line)
                except LogFormatError:
                    LOGGER.warning('line %s does not recognized as Phantom log')
                    continue
                if parsed.level not in ['error', 'fatal']:
                    continue
                counter[parsed.message] += 1
            return [err for (err, count) in
                    sorted(counter.items(), key=lambda item: -item[1])[:limit]]



from threading import Thread, Event

import requests
import time
import logging

logger = logging.getLogger(__name__)


class PandoraStatsPoller(Thread):
    def __init__(self, port):
        super(PandoraStatsPoller, self).__init__()
        self._stop_run = Event()
        self.buffer = []
        self.port = port

    def run(self):
        last_ts = int(time.time() - 1)

        while not self._stop_run.is_set():
            curr_ts = int(time.time())
            if curr_ts > last_ts:
                last_ts = curr_ts
                try:
                    pandora_stat = requests.get(
                        "http://localhost:{port}/debug/vars".format(port=self.port), timeout=0.9
                    ).json()
                    instances_metric = pandora_stat.get("engine_LastMaxActiveRequests", pandora_stat.get("engine_ActiveRequests"))
                    data = {
                        'ts': last_ts - 1,
                        'metrics': {
                            'instances': instances_metric,
                            'reqps': pandora_stat.get("engine_ReqPS"),
                        }
                    }
                except (requests.ConnectionError, requests.HTTPError, requests.exceptions.Timeout):
                    logger.debug("Pandora expvar http interface is unavailable", exc_info=True)
                    data = {
                        'ts': last_ts - 1,
                        'metrics': {
                            'instances': 0,
                            'reqps': 0
                        }
                    }
                self.buffer.append(data)
            else:
                time.sleep(0.2)

    def stop(self):
        self._stop_run.set()

    def get_data(self):
        result, self.buffer = self.buffer, []
        return result


class PandoraStatsReader(object):
    # TODO: maybe make stats collection asyncronous
    def __init__(self, expvar, port):
        self.closed = False
        self.expvar = expvar
        self.port = port
        self.poller = PandoraStatsPoller(port)
        self.started = False

    def __next__(self):
        if not self.expvar:
            if self.closed:
                raise StopIteration
            return [{
                'ts': int(time.time() - 1),
                'metrics': {
                    'instances': 0,
                    'reqps': 0
                }
            }]
        else:
            if self.closed:
                raise StopIteration()
            elif not self.started:
                self.poller.start()
                self.started = True
            return self.poller.get_data()

    def close(self):
        self.closed = True
        self.poller.stop()
        self.poller.join()

    def __iter__(self):
        return self



""" Autostop facility """
# pylint: disable=C0301
import logging
import os.path
import time
from datetime import datetime

from . import criterions as cr
from . import cumulative_criterions as cum_cr
from ..Console import Plugin as ConsolePlugin
from ...common.interfaces import AbstractPlugin, AggregateResultListener, AbstractInfoWidget, MonitoringDataListener

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    """ Plugin that accepts criterion classes and triggers autostop """
    SECTION = 'autostop'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        AggregateResultListener.__init__(self)
        MonitoringDataListener.__init__(self)

        self.cause_criterion = None
        self.imbalance_rps = 0
        self.imbalance_timestamp = 0
        self._criterions = {}
        self.custom_criterions = []
        self.counting = []
        self._stop_report_path = ''

    @staticmethod
    def get_key():
        return __file__

    def get_counting(self):
        """ get criterions that are activated """
        return self.counting

    def add_counting(self, obj):
        """ add criterion that activated """
        self.counting += [obj]

    def add_criterion_class(self, criterion_class):
        """ add new criterion class """
        self.custom_criterions += [criterion_class]

    def get_available_options(self):
        return ["autostop", "report_file"]

    def configure(self):
        aggregator = self.core.job.aggregator
        aggregator.add_result_listener(self)

        self._stop_report_path = os.path.join(
            self.core.artifacts_dir,
            self.get_option("report_file", 'autostop_report.txt'))

        self.add_criterion_class(cr.AvgTimeCriterion)
        self.add_criterion_class(cr.NetCodesCriterion)
        self.add_criterion_class(cr.HTTPCodesCriterion)
        self.add_criterion_class(cr.QuantileCriterion)
        self.add_criterion_class(cr.SteadyCumulativeQuantilesCriterion)
        self.add_criterion_class(cr.TimeLimitCriterion)
        self.add_criterion_class(cr.UsedInstancesCriterion)
        self.add_criterion_class(cum_cr.TotalFracTimeCriterion)
        self.add_criterion_class(cum_cr.TotalHTTPCodesCriterion)
        self.add_criterion_class(cum_cr.TotalNetCodesCriterion)
        self.add_criterion_class(cum_cr.TotalNegativeHTTPCodesCriterion)
        self.add_criterion_class(cum_cr.TotalNegativeNetCodesCriterion)
        self.add_criterion_class(cum_cr.TotalHTTPTrendCriterion)

    def prepare_test(self):
        criterions = self.get_option("autostop")
        for criterion_str in criterions:
            if not criterion_str:
                continue
            self.log.debug("Criterion string: %s", criterion_str)
            self._criterions[criterion_str] = self.__create_criterion(
                criterion_str)

        self.log.debug("Criterion objects: %s", self._criterions)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            self.log.debug(ex)
            console = None

        if console:
            console.add_info_widget(AutostopWidget(self))

    def start_test(self):
        for criterion in self._criterions.values():
            if isinstance(criterion, cr.TimeLimitCriterion):
                criterion.start_time = time.time()

    def is_test_finished(self):
        if self.cause_criterion:
            self.log.warning("Autostop criterion requested test stop: %s", self.cause_criterion.explain())
            return self.cause_criterion.get_rc()
        else:
            return -1

    def __create_criterion(self, criterion_str):
        """ instantiate criterion from config string """
        parsed = criterion_str.split("(")
        type_str = parsed[0].strip().lower()
        parsed[1] = parsed[1].split(")")[0].strip()

        for criterion_class in self.custom_criterions:
            if criterion_class.get_type_string() == type_str:
                return criterion_class(self, parsed[1])
        raise ValueError("Unsupported autostop criterion type: %s" % criterion_str)

    def on_aggregated_data(self, data, stat):
        self.counting = []
        if not self.cause_criterion:
            for criterion_text, criterion in self._criterions.items():
                if criterion.notify(data, stat):
                    self.cause_criterion = criterion
                    if self.cause_criterion.cause_second:
                        self.imbalance_rps = int(self.cause_criterion.cause_second[1]["metrics"]["reqps"])
                        if not self.imbalance_rps:
                            self.imbalance_rps = int(
                                self.cause_criterion.cause_second[0]["overall"]["interval_real"]["len"])
                        self.imbalance_timestamp = int(self.cause_criterion.cause_second[0].get('ts', datetime.utcnow().timestamp()))
                    self.core.publish('autostop', 'rps', self.imbalance_rps)
                    self.core.publish('autostop', 'reason', criterion.explain())
                    self.log.warning(
                        "Autostop criterion requested test stop on %d rps: %s", self.imbalance_rps, criterion_text)
                    with open(self._stop_report_path, 'w') as f:
                        f.write(criterion_text)
                    self.core.add_artifact_file(self._stop_report_path)

    def monitoring_data(self, data):
        for _, criterion in self._criterions.items():
            if isinstance(criterion, MonitoringDataListener):
                criterion.monitoring_data(data)


class AutostopWidget(AbstractInfoWidget):
    """ widget that displays counting criterions """

    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.owner = sender

    def get_index(self):
        return 25

    def render(self, screen):
        res = []
        candidates = self.owner.get_counting()
        for candidate in candidates:
            text, perc = candidate.widget_explain()
            if perc >= 0.95:
                res += [screen.markup.RED_DARK + text + screen.markup.RESET]
            elif perc >= 0.8:
                res += [screen.markup.RED + text + screen.markup.RESET]
            elif perc >= 0.5:
                res += [screen.markup.YELLOW + text + screen.markup.RESET]
            else:
                res += [text]

        if res:
            return "Autostop:\n  " + ("\n  ".join(res))
        else:
            return ''



from .local import LocalStorageClient
from .luna import LunaClient
from .lunapark_volta import LunaparkVoltaClient

available_clients = {
    'luna': LunaClient,
    'local_storage': LocalStorageClient,
    'lunapark_volta': LunaparkVoltaClient,
}



import ctypes
import logging
import os
import shutil
import time
from configparser import RawConfigParser, MissingSectionHeaderError
from multiprocessing import Event, Value, Process

import stat
import yaml

from yandextank.common.interfaces import TankInfo
from yandextank.common.util import Cleanup, Finish, Status, TankapiLogFilter, read_resource
from yandextank.config_converter.converter import convert_ini, convert_single_option
from yandextank.core import TankCore
from yandextank.core.tankcore import LockError, Lock
from yandextank.validator.validator import ValidationError

logger = logging.getLogger()


class TankWorker(Process):
    SECTION = 'core'
    FINISH_FILENAME = 'finish_status.yaml'
    DEFAULT_CONFIG = 'load.yaml'

    def __init__(self, configs, cli_options=None, cfg_patches=None, cli_args=None, no_local=False,
                 log_handlers=None, wait_lock=False, files=None, ammo_file=None, debug=False,
                 run_shooting_event=None, storage=None, resource_manager=None, plugins_implicit_enabling=False):
        super().__init__()
        self.interrupted = Event()
        self.info = TankInfo(dict())
        user_configs = self._combine_configs(configs, cli_options, cfg_patches, cli_args)
        self.core = TankCore(
            user_configs,
            self.interrupted,
            self.info,
            storage=storage,
            skip_base_cfgs=no_local,
            resource_manager=resource_manager,
            plugins_implicit_enabling=plugins_implicit_enabling,
        )

        is_locked = Lock.is_locked(self.core.lock_dir)
        if is_locked and not self.core.config.get_option(self.SECTION, 'ignore_lock'):
            raise LockError(is_locked)

        self.wait_lock = wait_lock
        self.log_handlers = log_handlers if log_handlers is not None else []
        self.files = [] if files is None else files
        self.ammo_file = ammo_file
        self.config_paths = configs
        self.folder = self.core.artifacts_dir
        self._logger_old_loglevel, self._logger_tank_handlers = self.init_logging(debug or self.core.get_option(self.core.SECTION, 'debug'))

        self._status = Value(ctypes.c_char_p, Status.TEST_INITIATED)
        self._test_id = Value(ctypes.c_char_p, self.core.test_id.encode('utf8'))
        self._retcode = Value(ctypes.c_int, 0)
        self._msgs = []
        self._run_shooting_event = run_shooting_event or self._dummy_event()

    @staticmethod
    def _dummy_event():
        event = Event()
        event.set()
        return event

    def run(self):
        with Cleanup(self) as add_cleanup:
            add_cleanup('cleanup log handlers', self.cleanup_logging)
            lock = self.get_lock()
            add_cleanup('release lock', lock.release)
            self.status = Status.TEST_PREPARING
            logger.info('Created a folder for the test. %s', self.folder)
            self.core.plugins_configure()
            add_cleanup('plugins cleanup', self.core.plugins_cleanup)
            self.core.plugins_prepare_test()
            with Finish(self):
                if not self._run_shooting_event.is_set():
                    self.status = Status.TEST_WAITING_FOR_A_COMMAND_TO_RUN
                    self._wait_for_a_command_to_start_shooting()
                self.status = Status.TEST_RUNNING
                self.core.plugins_start_test()
                self.retcode = self.core.wait_for_finish()
                self._msgs.extend(self.core.errors)
            self.status = Status.TEST_POST_PROCESS
            self.retcode = self.core.plugins_post_process(self.retcode)

    def _wait_for_a_command_to_start_shooting(self):
        pool_timeout = 0.01
        while True:
            if self.core.interrupted.is_set():
                raise RuntimeError('Test stopped before shooting started.')
            if self._run_shooting_event.wait(pool_timeout):
                return

    @staticmethod
    def _combine_configs(run_cfgs, cli_options, cfg_patches, cli_args):
        if cli_options is None:
            cli_options = []
        if cfg_patches is None:
            cfg_patches = []
        if cli_args is None:
            cli_args = []
        run_cfgs = run_cfgs if len(run_cfgs) > 0 else [TankWorker.DEFAULT_CONFIG]
        return (
            [load_cfg(cfg) for cfg in run_cfgs] +
            parse_options(cli_options) +
            parse_and_check_patches(cfg_patches) +
            cli_args
        )

    def stop(self):
        self.interrupted.set()
        logger.warning('Interrupting')

    def get_status(self):
        return {'status_code': self.status.decode('utf8'),
                'left_time': None,
                'exit_code': self.retcode,
                'lunapark_id': self.get_info('uploader', 'job_no'),
                'tank_msg': self.msg,
                'test_id': self.test_id,
                'lunapark_url': self.get_info('uploader', 'web_link')
                }

    def save_finish_status(self):
        with open(os.path.join(self.folder, self.FINISH_FILENAME), 'w') as f:
            status_data = self.get_status()
            status_data['status_code'] = Status.TEST_FINISHED.decode('utf8')
            yaml.safe_dump(status_data, f, encoding='utf-8', allow_unicode=True)

    def get_info(self, section_name, key_name):
        return self.info.get_value([section_name, key_name])

    def init_logging(self, debug=False):

        filename = os.path.join(self.core.artifacts_dir, 'tank.log')
        open(filename, 'a').close()
        current_file_mode = os.stat(filename).st_mode
        os.chmod(filename, current_file_mode | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

        handlers = []
        old_loglevel = logger.level
        logger.setLevel(logging.DEBUG if debug else logging.INFO)

        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)d\t%(message)s"))
        file_handler.addFilter(TankapiLogFilter())
        handlers.append(file_handler)
        logger.addHandler(file_handler)
        logger.info("Log file created")

        for handler in self.log_handlers:
            handlers.append(handler)
            logger.addHandler(handler)
            logger.info("Logging handler %s added", handler)
        return old_loglevel, handlers

    def cleanup_logging(self):
        if self._logger_tank_handlers is not None:
            for h in self._logger_tank_handlers:
                logger.removeHandler(h)
        if self._logger_old_loglevel is not None:
            logger.setLevel(self._logger_old_loglevel)

    def get_lock(self):
        while not self.interrupted.is_set():
            try:
                lock = Lock(self.test_id, self.folder).acquire(self.core.lock_dir,
                                                               self.core.config.get_option(self.SECTION, 'ignore_lock'))
                break
            except LockError as e:
                self.add_msgs(str(e))
                if not self.wait_lock:
                    raise RuntimeError("Lock file present, cannot continue")
                logger.warning("Couldn't get lock. Will retry in 5 seconds...")
                time.sleep(5)
        else:
            raise KeyboardInterrupt
        return lock

    @property
    def msg(self):
        return '\n'.join(self._msgs)

    def add_msgs(self, *msgs):
        self._msgs.extend(msgs)

    @property
    def test_id(self):
        with self._test_id.get_lock():
            return self._test_id.value.decode('utf8')

    @property
    def status(self):
        with self._status.get_lock():
            return self._status.value

    @status.setter
    def status(self, val):
        with self._status.get_lock():
            self._status.value = val

    @property
    def retcode(self):
        with self._retcode.get_lock():
            return self._retcode.value

    @retcode.setter
    def retcode(self, val):
        with self._retcode.get_lock():
            self._retcode.value = val

    def collect_files(self):
        for cfg in self.config_paths:
            shutil.move(cfg, self.folder)
        for f in self.files:
            shutil.move(f, self.folder)
        if self.ammo_file:
            shutil.move(self.ammo_file, self.folder)

    def go_to_test_folder(self):
        os.chdir(self.folder)


def load_cfg(cfg_filename):
    """
    :type cfg_filename: str
    """
    if is_ini(cfg_filename):
        return convert_ini(cfg_filename)
    else:
        cfg_yaml = yaml.load(read_resource(cfg_filename), Loader=yaml.FullLoader)
        if not isinstance(cfg_yaml, dict):
            raise ValidationError('Wrong config format, should be a yaml')
        return cfg_yaml


def parse_options(options):
    """
    :type options: list of str
    :rtype: list of dict
    """
    if options is None:
        return []
    else:
        return [
            convert_single_option(key.strip(), value.strip())
            for key, value
            in [option.split('=', 1) for option in options]
        ]


def parse_and_check_patches(patches):
    parsed = [yaml.load(p, Loader=yaml.FullLoader) for p in patches]
    for patch in parsed:
        if not isinstance(patch, dict):
            raise ValidationError('Config patch "{}" should be a dict'.format(patch))
    return parsed


def is_ini(cfg_file):
    if cfg_file.endswith('.yaml') or cfg_file.endswith('.json'):
        return False
    try:
        RawConfigParser().read(cfg_file)
        return True
    except MissingSectionHeaderError:
        return False



import logging
try:
    import queue as q
except ImportError:
    import Queue as q
import threading
import time


logger = logging.getLogger(__name__)


def get_nowait_from_queue(queue):
    """ Collect all immediately available items from a queue """
    data = []
    for _ in range(queue.qsize()):
        try:
            data.append(queue.get_nowait())
        except q.Empty:
            break
    return data


class Drain(threading.Thread):
    """
    Drain a generator to a destination that answers to put(), in a thread
    """

    def __init__(self, source, destination):
        super(Drain, self).__init__()
        self.source = source
        self.destination = destination
        self._finished = threading.Event()
        self._interrupted = threading.Event()
        self.daemon = True  # bdk+ytank stuck w/o this at join of this thread

    def run(self):
        try:
            for item in self.source:
                self.destination.put(item)
                if self._interrupted.is_set():
                    break
        except Exception as e:
            logger.error(e, exc_info=True)
            self._interrupted.set()
        finally:
            self._finished.set()

    def wait(self, timeout=None):
        self._finished.wait(timeout=timeout)

    def close(self):
        self._interrupted.set()


class Tee(threading.Thread):
    """Copy items from one queue to multiple in a thread.

    Note:
        Items are passed by reference.
    """

    def __init__(self, source, destination, type):
        """
        Args:
            source (queue): where to get items from
            destination (list): list of queues where to put items from the source
            type (string): ???
        """
        # TODO: what is type?!
        super(Tee, self).__init__()
        self.source = source
        self.destination = destination  # TODO: this is actually a list of destinations. Rename.
        self.type = type
        self._finished = threading.Event()
        self._interrupted = threading.Event()
        self.daemon = True  # just in case, bdk+ytank stuck w/o this at join of Drain thread

    def run(self):
        while not self._interrupted.is_set():
            data = get_nowait_from_queue(self.source)
            for item in data:
                for destination in self.destination:
                    destination.put(item, self.type)
                    if self._interrupted.is_set():
                        break
                if self._interrupted.is_set():
                    break
            if self._interrupted.is_set():
                break
            time.sleep(0.5)
        self._finished.set()

    def wait(self, timeout=None):
        self._finished.wait(timeout=timeout)

    def close(self):
        self._interrupted.set()


# TODO: does it really chop anything?
class Chopper(object):
    def __init__(self, source):
        self.source = source

    def __iter__(self):
        for chunk in self.source:
            for item in chunk:
                yield item



""" Resource Opener tool """
from __future__ import print_function

import logging
import os
import requests
import gzip
import hashlib
import serial
import yaml
import socket
import six
import typing
import environ
from yandextank.contrib.netort.netort.data_manager.common.util import thread_safe_property, YamlEnvSubstConfigLoader
from yandextank.contrib.netort.netort.data_manager.common.condition import uri_like, path_like, Condition
from urllib.parse import urlparse
from contextlib import closing
from dataclasses import dataclass
from functools import partial

logger = logging.getLogger(__name__)

try:
    import boto3
    import boto3.exceptions
    import boto3.s3
    import boto3.session
except ImportError:
    logger.debug('Failed to import `boto3` package. Install `boto3`, otherwise S3 file paths opener wont work')
    boto3 = None

try:
    from library.python import resource as rs
    pip = False
except ImportError:
    pip = True

_TMP_PATH_PREFIX = os.getenv('NETORT_TMP_PATH', '/tmp')


class PathProvider(object):
    def __init__(self, prefix) -> None:
        self.prefix = prefix

    def tmpfile_path(self, hash):
        hasher = hashlib.md5()
        hasher.update(six.ensure_binary(hash))
        return '%s_%s.downloaded_resource' % (self.prefix, hasher.hexdigest())


class FormatDetector(object):
    """ Format Detector
    """

    def __init__(self):
        self.formats = {'gzip': (0, b'\x1f\x8b'), 'tar': (257, b'ustar\x0000')}

    def detect_format(self, header):
        for fmt, signature in self.formats.items():
            if signature[1] == header[signature[0]:len(signature[1])]:
                return fmt


class OpenerProtocol(typing.Protocol):
    def __init__(self, path: str):
        pass

    def __call__(self):
        pass

    @property
    def get_filename(self):
        pass


class TempDownloaderOpenerProtocol(OpenerProtocol):
    def __call__(self, use_cache=True):
        pass

    def download_file(self, use_cache, try_ungzip=False):
        pass


@dataclass
class OpenerItem(object):
    condition: Condition
    factory: typing.Callable[[str], object]
    order: int = 0


OpenersConfig = typing.Collection[OpenerItem]


@environ.config(prefix='NETORT')
class ResourceManagerConfig(object):
    tmp_path = environ.var('/tmp')
    openers_config_path = environ.var('/etc/yandex-tank/netort_openers.config')
    s3_endpoint_url = environ.var('')
    aws_access_key_id = environ.var('')
    aws_secret_access_key = environ.var('')


class ResourceManager(object):
    """ Resource opener manager.
        Use resource_filename and resource_string methods.
    """

    def __init__(
        self,
        config: ResourceManagerConfig,
        openers: typing.Optional[OpenersConfig] = None,
    ):
        self.config = config
        logger.debug('loading ResourceManager config from %s', config.openers_config_path)
        self.openers_config = self.make_openers_config(config, self.load_config_safe(config.openers_config_path))

        self.tmp_path_prefix = config.tmp_path
        self.openers = openers or self._default_openers(config.tmp_path)

    def make_path_provider(self, subfolder: str) -> PathProvider:
        return PathProvider(os.path.join(self.tmp_path_prefix, subfolder))

    def _default_openers(self, tmp_path_prefix: str) -> OpenersConfig:
        http_opener = partial(HttpOpener, config=self.openers_config.get('http_opener'), path_provider=self.make_path_provider('http'))
        s3_opener = partial(S3Opener, config=self.openers_config.get('s3_opener'), path_provider=self.make_path_provider('s3'))

        return [
            OpenerItem(uri_like(scheme='http'), http_opener),
            OpenerItem(uri_like(scheme='https'), http_opener),
            OpenerItem(uri_like(scheme='s3'), s3_opener),
            OpenerItem(uri_like(scheme='file'), FileOpener),
            OpenerItem(path_like('/dev/'), SerialOpener),
        ]

    def make_openers_config(self, rm_config: ResourceManagerConfig, openers_config: typing.Optional[typing.Dict[str, typing.Any]]) -> typing.Dict[str, typing.Any]:
        openers_config = openers_config or {}
        if rm_config.aws_access_key_id and rm_config.aws_secret_access_key and rm_config.s3_endpoint_url:
            openers_config['s3_opener'] = {
                'aws_access_key_id': rm_config.aws_access_key_id,
                'aws_secret_access_key': rm_config.aws_secret_access_key,
                'endpoint_url': rm_config.s3_endpoint_url,
            }
        return openers_config

    def load_config_safe(self, path) -> typing.Optional[typing.Dict[str, typing.Any]]:
        if not path:
            return None
        if not os.path.exists(path):
            logger.info('Netort ResourceManager config file %s not exists.', path)
            return None
        try:
            with open(path, 'r') as f:
                return yaml.load(f, Loader=YamlEnvSubstConfigLoader)
        except Exception as e:
            logger.warning('Failed to load openers config file at %s: %s', path, str(e))
            return None

    def resource_filename(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            string, resource absolute path (downloads the url to /tmp)
        """
        return self.get_opener(path).get_filename

    def resource_string(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            string, file content
        """
        opener = self.get_opener(path)
        filename = opener.get_filename
        try:
            size = os.path.getsize(filename)
            if size > 50 * 1024 * 1024:
                logger.warning('Reading large resource to memory: %s. Size: %s bytes', filename, size)
        except Exception as exc:
            logger.debug('Unable to check resource size %s. %s', filename, exc)
        with opener(filename, 'r') as resource:
            content = resource.read()
        return content

    def get_opener(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            file object
        """
        path = rs.find(path) if not pip and path in rs.iterkeys(prefix='resfs/file/load/projects/yandex-tank/')\
            else path

        self._ensure_tmp_path_prefix_exists()
        openers = [o for o in self.openers if o.condition(path)]
        if not openers:
            return FileOpener(path)
        if len(openers) == 1:
            return openers[0].factory(path)

        explanation = '\n'.join([repr(o.condition) for o in openers])
        logger.info(f'multiple openers meets "{path}": {explanation}')
        opener = sorted(openers, key=lambda o: o.order, reverse=True)[0]
        logger.info(f'highest priority: {repr(opener.condition)}')
        return opener.factory(path)

    def _ensure_tmp_path_prefix_exists(self):
        if not os.path.exists(self.tmp_path_prefix):
            os.makedirs(self.tmp_path_prefix, exist_ok=True)


class SerialOpener(OpenerProtocol):
    """ Serial device opener.
    """

    def __init__(self, device, baud_rate=230400, read_timeout=1):
        self.baud_rate = baud_rate
        self.device = device
        self.read_timeout = read_timeout

    def __call__(self, *args, **kwargs):
        return serial.Serial(self.device, self.baud_rate, timeout=self.read_timeout)

    @property
    def get_filename(self):
        return self.device


class FileOpener(OpenerProtocol):
    """ File opener.
    """

    def __init__(self, f_path):
        self.f_path = f_path
        self.fmt_detector = FormatDetector()

    def __call__(self, *args, **kwargs):
        with open(self.f_path, 'rb') as resource:
            header = resource.read(300)
        fmt = self.fmt_detector.detect_format(header)
        logger.debug('Resource %s format detected: %s.', self.f_path, fmt)
        if fmt == 'gzip':
            return gzip.open(self.f_path, 'rb')
        else:
            return open(self.f_path, 'rb')

    @property
    def get_filename(self):
        return self.f_path

    @property
    def hash(self):
        hashed_str = os.path.realpath(self.f_path)
        stat = os.stat(self.f_path)
        cnt = 0
        for stat_option in stat:
            if cnt == 7:  # skip access time
                continue
            cnt += 1
            hashed_str += ";" + str(stat_option)
        hashed_str += ";" + str(os.path.getmtime(self.f_path))
        return hashed_str

    @property
    def data_length(self):
        return os.path.getsize(self.f_path)


def retry(func):
    def with_retry(self, *args, **kwargs):
        for i in range(self.attempts):
            try:
                return func(self, *args, **kwargs)
            except Exception:
                logger.exception('%s failed. Retrying.', func)
                continue
        return func(self, *args, **kwargs)
    return with_retry


class HttpOpener(TempDownloaderOpenerProtocol):
    """ Http url opener.
        Downloads small files.
        For large files returns wrapped http stream.
    """

    def __init__(self, url, timeout=5, attempts=5, path_provider=None, config=None):
        self._filename = None
        self.url = url
        self.fmt_detector = FormatDetector()
        self.force_download = None
        self.data_info = None
        self.timeout = timeout
        self.attempts = attempts
        self.path_provider = path_provider or PathProvider(os.path.join(_TMP_PATH_PREFIX, 'http'))
        self._use_config(url, config)
        self.get_request_info()

    def __call__(self, use_cache=True, *args, **kwargs):
        return self.open(use_cache, *args, **kwargs)

    def _use_config(self, url: str, config: typing.Optional[typing.Dict[str, typing.Any]]):
        self._default_headers = None
        if not config:
            return
        parsed = urlparse(self.url)
        host = parsed.netloc.rsplit(':')[0]
        config = config.get(host, {})
        self._default_headers = config.get('headers')

    @retry
    def open(self, use_cache, *args, **kwargs):
        with closing(
                requests.get(
                    self.url, stream=True, verify=False, headers=self._default_headers,
                    timeout=self.timeout)) as stream:
            stream_iterator = stream.raw.stream(100, decode_content=True)
            header = next(stream_iterator)
            while len(header) < 10:
                header += next(stream_iterator)
            fmt = self.fmt_detector.detect_format(header)
            logger.debug('Resource %s format detected: %s.', self.url, fmt)

        if not self.force_download and fmt != 'gzip' and self.data_length > 10**8:
            logger.info("Resource data is not gzipped and larger than 100MB. Reading from stream..")
            return HttpBytesStreamWrapper(self.url, headers=self._default_headers)
        else:
            downloaded_f_path = self.download_file(use_cache)
            if fmt == 'gzip':
                return gzip.open(downloaded_f_path, mode='rb')
            else:
                return open(downloaded_f_path, 'rb')

    @retry
    def download_file(self, use_cache, try_ungzip=False):
        tmpfile_path = self.tmpfile_path()
        if os.path.exists(tmpfile_path) and use_cache:
            logger.info("Resource %s has already been downloaded to %s . Using it..", self.url, tmpfile_path)
        else:
            logger.info("Downloading resource %s to %s", self.url, tmpfile_path)
            try:
                data = requests.get(self.url, verify=False, headers=self._default_headers, timeout=self.timeout)
                data.raise_for_status()
            except requests.exceptions.Timeout:
                logger.info('Connection timeout reached trying to download resource via HttpOpener: %s',
                            self.url, exc_info=True)
                raise
            except requests.exceptions.HTTPError:
                logger.error('Bad http code during resource downloading. Http code: %s, resource: %s', data.status_code, self.url)
                raise
            else:
                f = open(tmpfile_path, "wb")
                f.write(data.content)
                f.close()
                logger.info("Successfully downloaded resource %s to %s, http status code: %s", self.url, tmpfile_path, data.status_code)
        if try_ungzip:
            tmpfile_path = try_ungzip_file(tmpfile_path)
        self._filename = tmpfile_path
        return tmpfile_path

    def tmpfile_path(self):
        return self.path_provider.tmpfile_path(self.hash)

    @retry
    def get_request_info(self):
        logger.debug('Trying to get info about resource %s', self.url)
        headers = self._default_headers or {}
        headers.update({'Accept-Encoding': 'identity'})
        req = requests.Request(
            'HEAD', self.url, headers=headers)
        session = requests.Session()
        prepared = session.prepare_request(req)
        try:
            self.data_info = session.send(
                prepared,
                verify=False,
                allow_redirects=True,
                timeout=self.timeout)
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
            logger.warning('Connection error trying to get info for resource %s. Retrying...', self.url, exc_info=True)
            try:
                self.data_info = session.send(
                    prepared,
                    verify=False,
                    allow_redirects=True,
                    timeout=self.timeout)
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
                logger.warning(
                    'Connection error trying to get info for resource %s. Retrying...',
                    self.url, exc_info=True)
                raise
        finally:
            session.close()
        try:
            self.data_info.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            if exc.response.status_code == 405:
                logger.info(
                    "Resource storage does not support HEAD method. Ignore proto error and force download file."
                )
                self.force_download = True
            else:
                logger.warning('Invalid HTTP response trying to get info about resource: %s', self.url, exc_info=True)
                raise

    @thread_safe_property
    def get_filename(self):
        return self.download_file(use_cache=True)

    @property
    def hash(self):
        last_modified = self.data_info.headers.get("Last-Modified", '')
        hash = self.url + "|" + last_modified
        logger.info('Hash: {}'.format(hash))
        return self.url + "|" + last_modified

    @property
    def data_length(self):
        data_length = int(self.data_info.headers.get("Content-Length", 0))
        return data_length


class HttpBytesStreamWrapper:
    """
    makes http stream to look like file object
    """

    def __init__(self, url, headers=None):
        self.next = self.__next__
        self.url = url
        self.headers = headers
        self.buffer = b''
        self.pointer = 0
        self.stream_iterator = None
        self._content_consumed = False
        self.chunk_size = 10**3
        try:
            self.stream = requests.get(
                self.url, stream=True, verify=False, timeout=10, headers=headers)
            self.stream_iterator = self.stream.iter_content(self.chunk_size)
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
            logger.warning(
                'Connection errors or timeout reached trying to create HTTP stream for res: %s', self.url, exc_info=True
            )
            raise
        try:
            self.stream.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.warning('Invalid HTTP response trying to open stream for resource: %s', self.url, exc_info=True)
            raise

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.stream.connection.close()

    def __iter__(self):
        return self

    def _reopen_stream(self):
        self.stream.connection.close()
        try:
            self.stream = requests.get(
                self.url, stream=True, verify=False, timeout=30, headers=self.headers)
            self.stream_iterator = self.stream.iter_content(self.chunk_size)
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
            logger.warning(
                'Connection errors or timeout reached trying to reopen stream while downloading resource: %s',
                self.url, exc_info=True
            )
            raise
        try:
            self.stream.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.warning(
                'Invalid HTTP response trying to reopen stream for resource: %s',
                self.url, exc_info=True
            )
            raise
        self._content_consumed = False

    def _enhance_buffer(self):
        self.buffer += next(self.stream_iterator)

    def tell(self):
        return self.pointer

    def seek(self, position):
        if self.pointer:
            self.buffer = b''
            self._reopen_stream()
            self._enhance_buffer()
            while len(self.buffer) < position:
                self._enhance_buffer()
            self.pointer = position
            self.buffer = self.buffer[position:]

    def __next__(self):
        while b'\n' not in self.buffer:
            try:
                self._enhance_buffer()
            except (
                    StopIteration, TypeError,
                    requests.exceptions.StreamConsumedError):
                self._content_consumed = True
                break
        if not self._content_consumed or self.buffer:
            try:
                line = self.buffer[:self.buffer.index(b'\n') + 1]
            except ValueError:
                line = self.buffer
            self.pointer += len(line)
            self.buffer = self.buffer[len(line):]
            return line
        raise StopIteration

    def read(self, chunk_size):
        while len(self.buffer) < chunk_size:
            try:
                self._enhance_buffer()
            except (
                    StopIteration, TypeError,
                    requests.exceptions.StreamConsumedError):
                break
        if len(self.buffer) > chunk_size:
            chunk = self.buffer[:chunk_size]
        else:
            chunk = self.buffer
        self.pointer += len(chunk)
        self.buffer = self.buffer[len(chunk):]
        return chunk

    def readline(self):
        """
        requests iter_lines() uses splitlines() thus losing '\r\n'
        we need a different behavior for AmmoFileReader
        and we have to use our buffer because we have probably read
        a bunch into it already
        """
        try:
            return next(self)
        except StopIteration:
            return b''


class S3Opener(TempDownloaderOpenerProtocol):
    """ Simple Storage Service opener
        Downloads files.

        s3credentials.json fmt:
        {
            "aws_access_key_id": "key-id",
            "aws_secret_access_key": "secret-id",
            "host": "hostname.tld",
            "port": 7480,
            "is_secure": false
        }
    """

    def __init__(self, uri, credentials_path='/etc/yandex-tank/s3credentials.json', path_provider=None, config=None):
        s3_credentials = config
        if not s3_credentials:
            with open(credentials_path) as fname:
                s3_credentials = yaml.load(fname.read())
        self.endpoint_url = s3_credentials.get('endpoint_url')
        self.aws_access_key_id = s3_credentials.get('aws_access_key_id')
        self.aws_secret_access_key = s3_credentials.get('aws_secret_access_key')
        self.uri = uri
        urlparsed = urlparse(self.uri)
        self.bucket_key = urlparsed.netloc
        self.object_key = urlparsed.path.strip('/')
        self.path_provider = path_provider or PathProvider(os.path.join(_TMP_PATH_PREFIX, 's3'))
        self._filename = None
        self._conn = None

    @thread_safe_property
    def conn(self):
        if self._conn is None:
            if not boto3:
                raise RuntimeError("Install 'boto3' python package manually please")
            logger.debug('Opening connection to s3 %s', self.endpoint_url)
            session = boto3.session.Session()
            self._conn = session.client(
                service_name='s3',
                endpoint_url=self.endpoint_url,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
            )
        return self._conn

    def __call__(self, *args, **kwargs):
        return self.open(*args, **kwargs)

    def open(self):
        if not self._filename:
            self._filename = self.download_file(True, try_ungzip=False)
        return open(self._filename, 'rb')

    def tmpfile_path(self):
        return self.path_provider.tmpfile_path(self.hash)

    def download_file(self, use_cache, try_ungzip=False):
        tmpfile_path = self.tmpfile_path()
        if os.path.exists(tmpfile_path) and use_cache:
            logger.info("Resource %s has already been downloaded to %s . Using it..", self.uri, tmpfile_path)
        else:
            if not self.conn:
                raise Exception('Connection should be initialized first')
            logger.info("Downloading resource %s to %s", self.uri, tmpfile_path)

            try:
                self.conn.download_file(self.bucket_key, self.object_key, tmpfile_path)
            except socket.gaierror:
                logger.error('Failed to connect to s3 host %s', self.endpoint_url)
                raise
            except boto3.exceptions.Boto3Error as e:
                logger.error('S3 error trying to download file from bucket: %s/%s  %s', self.bucket_key,
                             self.object_key, str(e))
                logger.debug('S3 error trying to download file from bucket: %s/%s', self.bucket_key, self.object_key,
                             exc_info=True)
                raise
            except Exception as e:
                logger.debug('Failed to get s3 resource: %s', self.uri, exc_info=True)
                raise RuntimeError('Failed to get s3 resource %s' % self.uri) from e

            logger.info("Successfully downloaded resource %s to %s", self.uri, tmpfile_path)

        if try_ungzip:
            tmpfile_path = try_ungzip_file(tmpfile_path)
        self._filename = tmpfile_path
        return tmpfile_path

    @property
    def hash(self):
        hashed_str = "{bucket_key}_{object_key}".format(
            bucket_key=self.bucket_key,
            object_key=self.object_key,
        )
        return hashed_str

    @property
    def data_length(self):
        return os.path.getsize(self.get_filename)


def make_resource_manager():
    return ResourceManager(environ.to_config(ResourceManagerConfig))


manager = make_resource_manager()


def try_ungzip_file(file_path: str) -> str:
    try:
        if file_path.endswith('.gz'):
            ungzippedfile_path = file_path[:-3]
        else:
            ungzippedfile_path = file_path + '_ungzipped'
        with gzip.open(file_path) as gzf, open(ungzippedfile_path, 'wb') as f:
            f.write(gzf.read())
        return ungzippedfile_path
    except IOError as ioe:
        logger.error('Failed trying to unzip downloaded resource %s' % repr(ioe))
    return file_path



import fnmatch
import typing
from urllib.parse import urlparse


class Condition(object):
    def __init__(self, callable_: typing.Callable, explanation: str):
        self._callable = callable_
        self._explanation = explanation

    def __call__(self, *args, **kwargs):
        return self._callable(*args, **kwargs)

    def __repr__(self) -> str:
        return self._explanation

    def __str__(self) -> str:
        return self.__repr__()


def uri_like(
    *,
    scheme: typing.Optional[str] = None,
    host: typing.Optional[str] = None,
    path: typing.Optional[str] = None,
) -> Condition:
    if all(arg is None for arg in [scheme, host, path]):
        raise ValueError('uri_like: at least one argument must be specified')

    def condition(url: str, *args, **kwargs):
        parsed = urlparse(url)
        meet = True
        if scheme is not None:
            meet = meet and parsed.scheme.lower() == scheme.lower()
        if host is not None:
            meet = meet and _host_match(parsed.netloc, host)
        if path is not None:
            meet = meet and fnmatch.fnmatch(parsed.path.lower(), path.lower())
        return meet

    explanation = '/'.join(filter(None, [
        f'{scheme}:/' if scheme is not None else 'scheme:/',
        host if host is not None else '*',
        path,
    ]))
    return Condition(condition, f'uri like {explanation}')


def and_(*conditions) -> Condition:
    explanation = ' and '.join([repr(c) for c in conditions])
    return Condition(lambda *args, **kwargs: all([c(*args, **kwargs) for c in conditions]), explanation)


def path_like(pattern: str) -> Condition:
    def condition(path: str, *args, **kwargs):
        return fnmatch.fnmatchcase(path, pattern)
    return Condition(condition, f'path_like "{pattern}"')


def _host_match(netloc: str, pattern: str) -> bool:
    netloc_host = netloc.rsplit(':', 1)[0]
    return fnmatch.fnmatch(netloc_host, pattern)



import datetime
import logging
import os
import socket
import time

from ...common.interfaces import AbstractInfoWidget

from ..Console import screen

logger = logging.getLogger(__name__)


class PhantomProgressBarWidget(AbstractInfoWidget):
    """
    Widget that displays progressbar
    """

    def get_index(self):
        return 0

    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.krutilka = screen.krutilka()
        self.owner = sender
        self.ammo_progress = 0
        self.eta_file = None

        info = self.owner.get_info()
        if info:
            self.ammo_count = int(info.ammo_count)
            self.test_duration = int(info.duration)
        else:
            self.ammo_count = 1
            self.test_duration = 1

    def render(self, screen):
        res = ""

        dur_seconds = int(time.time()) - int(self.owner.start_time)

        eta_time = 'N/A'
        eta_secs = -1
        progress = 0
        color_bg = screen.markup.BG_CYAN
        color_fg = screen.markup.CYAN
        if self.test_duration and self.test_duration >= dur_seconds:
            color_bg = screen.markup.BG_GREEN
            color_fg = screen.markup.GREEN
            eta_secs = self.test_duration - dur_seconds
            eta_time = datetime.timedelta(seconds=eta_secs)
            progress = float(dur_seconds) / self.test_duration
        elif self.ammo_progress:
            left_part = self.ammo_count - self.ammo_progress
            if left_part > 0:
                eta_secs = int(
                    float(dur_seconds) / float(self.ammo_progress) * float(left_part))
            else:
                eta_secs = 0
            eta_time = datetime.timedelta(seconds=eta_secs)
            if self.ammo_progress < self.ammo_count:
                progress = float(self.ammo_progress) / float(self.ammo_count)
            else:
                progress = 0.5

        if self.eta_file:
            handle = open(self.eta_file, 'w')
            handle.write(str(eta_secs))
            handle.close()

        perc = float(int(1000 * progress)) / 10
        str_perc = str(perc) + "%"

        pb_width = screen.right_panel_width - 1 - len(str_perc)

        progress_chars = '=' * (int(pb_width * progress) - 1)
        progress_chars += next(self.krutilka)

        res += color_bg + progress_chars + screen.markup.RESET + color_fg
        res += '~' * (pb_width - int(pb_width * progress)
                      ) + screen.markup.RESET + ' '
        res += str_perc + "\n"

        eta = 'ETA: %s' % eta_time
        dur = 'Duration: %s' % str(datetime.timedelta(seconds=dur_seconds))
        spaces = ' ' * (screen.right_panel_width - len(eta) - len(dur) - 1)
        res += dur + ' ' + spaces + eta

        return res

    def on_aggregated_data(self, data, stats):
        self.ammo_progress += data["overall"]["interval_real"]["len"]


class PhantomInfoWidget(AbstractInfoWidget):
    """
    Widget with information about current run state
    """

    def get_index(self):
        return 2

    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.owner = sender
        self.instances = 0
        self.planned = 0
        self.RPS = 0
        self.selfload = 0
        self.time_lag = 0
        self.planned_rps_duration = 0

        info = self.owner.get_info()
        if info:
            self.instances_limit = int(info.instances)
            self.ammo_count = int(info.ammo_count)
        else:
            self.instances_limit = 1
            self.ammo_count = 1

    def render(self, screen):
        res = ''
        info = self.owner.get_info()
        if self.owner.phantom:
            template = "Hosts: %s => %s:%s\n Ammo: %s\nCount: %s\n Load: %s"
            data = (
                socket.gethostname(), info.address, info.port,
                os.path.basename(info.ammo_file), self.ammo_count,
                ' '.join(info.rps_schedule))
            res = template % data

            res += "\n\n"

        res += "Active instances: "
        if float(self.instances) / self.instances_limit > 0.8:
            res += screen.markup.RED + str(self.instances) + screen.markup.RESET
        elif float(self.instances) / self.instances_limit > 0.5:
            res += screen.markup.YELLOW + str(
                self.instances) + screen.markup.RESET
        else:
            res += str(self.instances)

        res += "\nPlanned requests: %s for %s\nActual responses: " % (
            self.planned, datetime.timedelta(seconds=self.planned_rps_duration))
        if self.planned != self.RPS:
            res += screen.markup.YELLOW + str(self.RPS) + screen.markup.RESET
        else:
            res += str(self.RPS)

        res += "\n        Accuracy: "
        if self.selfload < 80:
            res += screen.markup.RED + (
                '%.2f' % self.selfload) + screen.markup.RESET
        elif self.selfload < 95:
            res += screen.markup.YELLOW + (
                '%.2f' % self.selfload) + screen.markup.RESET
        else:
            res += ('%.2f' % self.selfload)

        res += "%\n        Time lag: "
        if self.time_lag > self.owner.buffered_seconds * 5:
            logger.debug("Time lag: %s", self.time_lag)
            res += screen.markup.RED + str(
                datetime.timedelta(seconds=self.time_lag)) + screen.markup.RESET
        elif self.time_lag > self.owner.buffered_seconds:
            res += screen.markup.YELLOW + str(
                datetime.timedelta(seconds=self.time_lag)) + screen.markup.RESET
        else:
            res += str(datetime.timedelta(seconds=self.time_lag))

        return res

    def on_aggregated_data(self, data, stats):
        self.RPS = data["overall"]["interval_real"]["len"]
        self.planned = stats["metrics"]["reqps"]
        self.instances = stats["metrics"]["instances"]


#           TODO:
#           self.selfload = second_aggregate_data.overall.selfload
#           self.time_lag = int(time.time() - time.mktime(
#               second_aggregate_data.time.timetuple()))



import json
import time
import traceback
import urllib.parse
import uuid
import requests
import logging

from requests.exceptions import ConnectionError, Timeout
from urllib3.exceptions import ProtocolError


requests.packages.urllib3.disable_warnings()
logger = logging.getLogger(__name__)  # pylint: disable=C0103


def id_gen(base, start=0):
    i = start
    while True:
        yield '%s-%d' % (base, i)
        i += 1


class APIClient(object):
    REQUEST_ID_HEADER = 'X-Request-ID'

    def __init__(
            self,
            core_interrupted,
            base_url=None,
            writer_url=None,
            network_attempts=10,
            api_attempts=10,
            maintenance_attempts=40,
            network_timeout=2,
            api_timeout=5,
            maintenance_timeout=15,
            connection_timeout=5.0,
            user_agent=None,
            api_token=None):
        self.core_interrupted = core_interrupted
        self.user_agent = user_agent
        self.connection_timeout = connection_timeout
        self._base_url = base_url
        self.writer_url = writer_url

        self.retry_timeout = 10
        self.session = requests.Session()
        self.session.verify = False
        self.session.headers.update({"User-Agent": "tank"})

        if "https" in requests.utils.getproxies():
            logger.info("Connecting via proxy %s" % requests.utils.getproxies()['https'])
            self.session.proxies = requests.utils.getproxies()
        else:
            logger.info("Proxy not set")

        self.network_attempts = network_attempts
        self.network_timeout = network_timeout
        self.api_attempts = api_attempts
        self.api_timeout = api_timeout
        self.maintenance_attempts = maintenance_attempts
        self.maintenance_timeout = maintenance_timeout
        self.params = {'api_token': api_token} if api_token else {}

    @property
    def base_url(self):
        if not self._base_url:
            raise ValueError("Base url is not set")
        else:
            return self._base_url

    @base_url.setter
    def base_url(self, url):
        self._base_url = url

    class UnderMaintenance(Exception):
        message = "API is under maintenance"

    class NotAvailable(Exception):
        desc = "API is not available"

        def __init__(self, request, response):
            self.message = "%s\n%s\n%s" % (self.desc, request, response)
            super(self.__class__, self).__init__(self.message)

    class StoppedFromOnline(Exception):
        """http code 410"""
        message = "Shooting is stopped from online"

    class JobNotCreated(Exception):
        pass

    class NetworkError(Exception):
        pass

    def set_api_timeout(self, timeout):
        self.api_timeout = float(timeout)

    def network_timeouts(self):
        return (self.network_timeout for _ in range(self.network_attempts - 1))

    def api_timeouts(self):
        return (self.api_timeout for _ in range(self.api_attempts - 1))

    def maintenance_timeouts(self):
        return (
            self.maintenance_timeout for _ in range(
                self.maintenance_attempts - 1))

    @staticmethod
    def filter_headers(headers):
        boring = ['X-Content-Security-Policy', 'Content-Security-Policy',
                  'Strict-Transport-Security', 'X-WebKit-CSP', 'Set-Cookie',
                  'X-DNS-Prefetch-Control', 'X-Frame-Options', 'P3P',
                  'X-Content-Type-Options', 'X-Download-Options',
                  'Surrogate-Control']
        for h in boring:
            if h in headers:
                del (headers[h])
        return headers

    def __send_single_request(self, request, request_id, trace=False):
        request.headers[self.REQUEST_ID_HEADER] = request_id
        p = self.session.prepare_request(request)
        if trace:
            logger.debug(self.format_request_info(p, request_id))
        resp = self.session.send(p, timeout=self.connection_timeout)
        if trace:
            logger.debug(self.format_response_info(resp, request_id))
        if resp.status_code in [500, 502, 503, 504]:
            raise self.NotAvailable(
                request=self.format_request_info(p, request_id),
                response=self.format_response_info(resp, request_id))
        elif resp.status_code == 410:
            raise self.StoppedFromOnline
        elif resp.status_code == 423:
            raise self.UnderMaintenance
        else:
            resp.raise_for_status()
            return resp

    def format_request_info(self, request, request_id):
        utf8_body = request.body.decode('utf-8') if isinstance(request.body, bytes) else request.body
        request_info = {
            'id': request_id,
            'method': request.method,
            'url': request.url,
            'headers': str(self.filter_headers(request.headers)),
            'body': utf8_body.replace('\n', '\\n') if isinstance(utf8_body, str) else utf8_body
        }
        return """Request: {}""".format(json.dumps(request_info))

    def format_response_info(self, resp, request_id):
        response_info = {
            'id': request_id,
            'elapsed_time': resp.elapsed.total_seconds(),
            'reason': resp.reason,
            'http code': resp.status_code,
            'headers': str(self.filter_headers(resp.headers)),
            'content': resp.text.replace('\n', '\\n') if isinstance(resp.text, str) else resp.text
        }
        return """Response: {}""".format(json.dumps(response_info))

    def __make_api_request(
            self,
            http_method,
            path,
            data=None,
            response_callback=lambda x: x,
            writer=False,
            interrupted_event=None,
            trace=False,
            json=None,
            maintenance_timeouts=None,
            maintenance_msg=None):
        url = urllib.parse.urljoin(self.base_url, path)
        ids = id_gen(str(uuid.uuid4()))
        if json:
            request = requests.Request(
                http_method, url, json=json, headers={'User-Agent': self.user_agent}, params=self.params)
        else:
            request = requests.Request(
                http_method, url, data=data, headers={'User-Agent': self.user_agent}, params=self.params)
        network_timeouts = self.network_timeouts()
        maintenance_timeouts = maintenance_timeouts or self.maintenance_timeouts()
        maintenance_msg = maintenance_msg or "%s is under maintenance" % (self._base_url)
        while interrupted_event is None or not interrupted_event.is_set():
            try:
                response = self.__send_single_request(request, next(ids), trace=trace)
                return response_callback(response)
            except (Timeout, ConnectionError, ProtocolError):
                logger.warn(traceback.format_exc())
                if not self.core_interrupted.is_set():
                    try:
                        timeout = next(network_timeouts)
                        logger.warning("Network error, will retry in %ss..." % timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise self.NetworkError()
                else:
                    break
            except self.UnderMaintenance as e:
                if not self.core_interrupted.is_set():
                    try:
                        timeout = next(maintenance_timeouts)
                        logger.warning(maintenance_msg)
                        logger.warning("Retrying in %ss..." % timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise e
                else:
                    break

    def __make_writer_request(
            self,
            params=None,
            json=None,
            http_method="POST",
            trace=False):
        '''
        Send request to writer service.
        '''
        request = requests.Request(
            http_method,
            self.writer_url,
            params=params,
            json=json,
            headers={
                'User-Agent': self.user_agent})
        ids = id_gen(str(uuid.uuid4()))
        network_timeouts = self.network_timeouts()
        maintenance_timeouts = self.maintenance_timeouts()
        while True:
            try:
                response = self.__send_single_request(request, next(ids), trace=trace)
                return response
            except (Timeout, ConnectionError, ProtocolError):
                logger.warning(traceback.format_exc())
                try:
                    timeout = next(network_timeouts)
                    logger.warning("Network error, will retry in %ss..." % timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise self.NetworkError()
            except self.UnderMaintenance as e:
                try:
                    timeout = next(maintenance_timeouts)
                    logger.warning("Writer is under maintenance, will retry in %ss..." % timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def __get(self, addr, trace=False, maintenance_timeouts=None, maintenance_msg=None):
        return self.__make_api_request(
            'GET',
            addr,
            trace=trace,
            response_callback=lambda r: json.loads(r.content.decode('utf8')),
            maintenance_timeouts=maintenance_timeouts,
            maintenance_msg=maintenance_msg
        )

    def __post_raw(self, addr, txt_data, trace=False, interrupted_event=None):
        return self.__make_api_request(
            'POST', addr, txt_data, lambda r: r.content, trace=trace, interrupted_event=interrupted_event)

    def __post(self, addr, data, interrupted_event=None, trace=False):
        return self.__make_api_request(
            'POST',
            addr,
            json=data,
            response_callback=lambda r: r.json(),
            interrupted_event=interrupted_event,
            trace=trace)

    def __put(self, addr, data, trace=False):
        return self.__make_api_request(
            'PUT',
            addr,
            json=data,
            response_callback=lambda r: r.text,
            trace=trace)

    def __patch(self, addr, data, trace=False):
        return self.__make_api_request(
            'PATCH',
            addr,
            json=data,
            response_callback=lambda r: r.text,
            trace=trace)

    def get_task_data(self, task, trace=False):
        return self.__get("api/task/" + task + "/summary.json", trace=trace)

    def new_job(
            self,
            task,
            person,
            tank,
            target_host,
            target_port,
            loadscheme=None,
            detailed_time=None,
            notify_list=None,
            trace=False):
        """
        :return: job_nr, upload_token
        :rtype: tuple
        """
        if not notify_list:
            notify_list = []
        data = {
            'task': task,
            'person': person,
            'tank': tank,
            'host': target_host,
            'port': target_port,
            'loadscheme': loadscheme,
            'detailed_time': detailed_time,
            'notify': notify_list
        }

        logger.debug("Job create request: %s", data)
        api_timeouts = self.api_timeouts()
        while True:
            try:
                response = self.__post(
                    "api/job/create.json", data, trace=trace)[0]
                # [{"upload_token": "1864a3b2547d40f19b5012eb038be6f6", "job": 904317}]
                return response['job'], response['upload_token']
            except (self.NotAvailable, self.StoppedFromOnline) as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warning("API error, will retry in %ss..." % timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    logger.warning('Failed to create job on lunapark')
                    raise self.JobNotCreated() from e
            except requests.HTTPError as e:
                raise self.JobNotCreated('Failed to create job on lunapark\n{}'.format(e.response.content))
            except Exception:
                logger.exception('Failed to create job on lunapark')
                raise self.JobNotCreated()

    def get_job_summary(self, jobno):
        result = self.__get('api/job/' + str(jobno) + '/summary.json')
        return result[0]

    def close_job(self, jobno, retcode, trace=False):
        params = {'exitcode': str(retcode)}

        result = self.__get('api/job/' + str(jobno) + '/close.json?'
                            + urllib.parse.urlencode(params), trace=trace)
        return result[0]['success']

    def edit_job_metainfo(
            self,
            jobno,
            job_name,
            job_dsc,
            instances,
            ammo_path,
            loop_count,
            version_tested,
            component,
            cmdline,
            is_starred,
            tank_type=0,
            trace=False):
        data = {
            'name': job_name,
            'description': job_dsc,
            'instances': str(instances),
            'ammo': ammo_path,
            'loop': loop_count,
            'version': version_tested,
            'component': component,
            'tank_type': int(tank_type),
            'command_line': cmdline,
            'starred': int(is_starred)
        }

        response = self.__post(
            'api/job/' + str(jobno) + '/edit.json',
            data,
            trace=trace)
        return response

    def set_imbalance_and_dsc(self, jobno, rps, comment, ts=0):
        data = {}
        if rps:
            data['imbalance'] = rps
        if comment:
            res = self.get_job_summary(jobno)
            data['description'] = (res['dsc'] + "\n" + comment).strip()

        response = self.__post('api/job/' + str(jobno) + '/edit.json', data)
        return response

    @staticmethod
    def second_data_to_push_item(data, stat, timestamp, overall, case):
        """
        @data: SecondAggregateDataItem
        """
        api_data = {
            'overall': overall,
            'case': case,
            'net_codes': [],
            'http_codes': [],
            'time_intervals': [],
            'trail': {
                'time': str(timestamp),
                'reqps': stat["metrics"]["reqps"],
                'resps': data["interval_real"]["len"],
                'expect': data["interval_real"]["total"] / 1000.0 / data["interval_real"]["len"],
                'disper': 0,
                'self_load':
                    0,  # TODO abs(round(100 - float(data.selfload), 2)),
                'input': data["size_in"]["total"],
                'output': data["size_out"]["total"],
                'connect_time': data["connect_time"]["total"] / 1000.0 / data["connect_time"]["len"],
                'send_time':
                    data["send_time"]["total"] / 1000.0 / data["send_time"]["len"],
                'latency':
                    data["latency"]["total"] / 1000.0 / data["latency"]["len"],
                'receive_time': data["receive_time"]["total"] / 1000.0 / data["receive_time"]["len"],
                'threads': stat["metrics"]["instances"],  # TODO
            }
        }

        for q, value in zip(data["interval_real"]["q"]["q"],
                            data["interval_real"]["q"]["value"]):
            api_data['trail']['q' + str(q)] = value / 1000.0

        for code, cnt in data["net_code"]["count"].items():
            api_data['net_codes'].append({'code': int(code),
                                          'count': int(cnt)})

        for code, cnt in data["proto_code"]["count"].items():
            api_data['http_codes'].append({'code': int(code),
                                           'count': int(cnt)})

        api_data['time_intervals'] = APIClient.convert_hist(data["interval_real"][
            "hist"])
        return api_data

    @staticmethod
    def convert_hist(hist):
        data = hist['data']
        bins = hist['bins']
        return [
            {
                "from": 0,  # deprecated
                "to": b / 1000.0,
                "count": count,
            } for b, count in zip(bins, data)
        ]

    def push_test_data(
            self,
            jobno,
            upload_token,
            data_item,
            stat_item,
            interrupted_event,
            trace=False):
        items = []
        uri = 'api/job/{0}/push_data.json?upload_token={1}'.format(
            jobno, upload_token)
        ts = data_item["ts"]
        for case_name, case_data in data_item["tagged"].items():
            if case_name == "":
                case_name = "__NOTAG__"
            push_item = self.second_data_to_push_item(case_data, stat_item, ts,
                                                      0, case_name)
            items.append(push_item)
        overall = self.second_data_to_push_item(data_item["overall"],
                                                stat_item, ts, 1, '')
        items.append(overall)

        api_timeouts = self.api_timeouts()
        while not interrupted_event.is_set():
            try:
                if self.writer_url:
                    res = self.__make_writer_request(
                        params={
                            "jobno": jobno,
                            "upload_token": upload_token,
                        },
                        json={
                            "trail": items,
                        },
                        trace=trace)
                    logger.debug("Writer response: %s", res.text)
                    return res.json()["success"]
                else:
                    res = self.__post(uri, items, interrupted_event, trace=trace)
                    logger.debug("API response: %s", res)
                    success = int(res[0]['success'])
                    return success
            except self.NotAvailable as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warning("API error, will retry in %ss...", timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def push_monitoring_data(
            self,
            jobno,
            upload_token,
            send_data,
            interrupted_event,
            trace=False):
        if send_data:
            addr = "api/monitoring/receiver/push?job_id=%s&upload_token=%s" % (
                jobno, upload_token)
            api_timeouts = self.api_timeouts()
            while not interrupted_event.is_set():
                try:
                    if self.writer_url:
                        res = self.__make_writer_request(
                            params={
                                "jobno": jobno,
                                "upload_token": upload_token,
                            },
                            json={
                                "monitoring": send_data,
                            },
                            trace=trace)
                        logger.debug("Writer response: %s", res.text)
                        return res.json()["success"]
                    else:
                        res = self.__post_raw(
                            addr, json.dumps(send_data), trace=trace, interrupted_event=interrupted_event)
                        logger.debug("API response: %s", res)
                        success = res == 'ok'
                        return success
                except self.NotAvailable as e:
                    try:
                        timeout = next(api_timeouts)
                        logger.warning("API error, will retry in %ss...", timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise e

    def push_events_data(self, jobno, operator, send_data):
        if send_data:
            # logger.info('send data: %s', send_data)
            for key in send_data:
                addr = "/api/job/{jobno}/event.json".format(
                    jobno=jobno,
                )
                body = dict(
                    operator=operator,
                    text=key[1],
                    timestamp=key[0]
                )
                api_timeouts = self.api_timeouts()
                while True:
                    try:
                        # logger.debug('Sending event: %s', body)
                        res = self.__post_raw(addr, body)
                        logger.debug("API response for events push: %s", res)
                        success = res == 'ok'
                        return success
                    except self.NotAvailable as e:
                        try:
                            timeout = next(api_timeouts)
                            logger.warning("API error, will retry in %ss...", timeout)
                            time.sleep(timeout)
                            continue
                        except StopIteration:
                            raise e

    def send_status(self, jobno, upload_token, status, trace=False):
        addr = "api/v2/jobs/%s/?upload_token=%s" % (jobno, upload_token)
        status_line = status.get("core", {}).get("stage", "unknown")
        if "stepper" in status:
            status_line += " %s" % status["stepper"].get("progress")
        api_timeouts = self.api_timeouts()
        while True:
            try:
                self.__patch(addr, {"status": status_line}, trace=trace)
                return
            except self.NotAvailable as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warning("API error, will retry in %ss...", timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def is_target_locked(self, target, trace=False):
        addr = "api/server/lock.json?action=check&address=%s" % target
        res = self.__get(addr, trace=trace)
        return res[0]

    def lock_target(self, target, duration, trace=False, maintenance_timeouts=None, maintenance_msg=None):
        addr = "api/server/lock.json?action=lock&" + \
               "address=%s&duration=%s&jobno=None" % \
               (target, int(duration))
        res = self.__get(addr, trace=trace, maintenance_timeouts=maintenance_timeouts, maintenance_msg=maintenance_msg)
        return res[0]

    def unlock_target(self, target):
        addr = self.get_manual_unlock_link(target)
        res = self.__get(addr)
        return res[0]

    def get_virtual_host_info(self, hostname):
        addr = "api/server/virtual_host.json?hostname=%s" % hostname
        res = self.__get(addr)
        try:
            return res[0]
        except KeyError:
            raise Exception(res['error'])

    @staticmethod
    def get_manual_unlock_link(target):
        return "api/server/lock.json?action=unlock&address=%s" % target

    def send_config(self, jobno, lp_requisites, config_content, trace=False):
        endpoint, field_name = lp_requisites
        logger.debug("Sending {} config".format(field_name))
        addr = "/api/job/%s/%s" % (jobno, endpoint)
        self.__post_raw(addr, {field_name: config_content}, trace=trace)

    def link_mobile_job(self, lp_key, mobile_key):
        addr = "/api/job/{jobno}/edit.json".format(jobno=lp_key)
        data = {
            'mobile_key': mobile_key
        }
        response = self.__post(addr, data)
        return response


class LPRequisites:
    CONFIGINFO = ('configinfo.txt', 'configinfo')
    MONITORING = ('jobmonitoringconfig.txt', 'monitoringconfig')
    CONFIGINITIAL = ('configinitial.txt', 'configinitial')


class OverloadClient(APIClient):
    """ mocks below for nonexistent backend methods """
    def send_status(self, jobno, upload_token, status, trace=False):
        return

    def lock_target(self, target, duration, trace=False, **kwargs):
        return

    def unlock_target(self, *args, **kwargs):
        return

    def link_mobile_job(self, lp_key, mobile_key):
        return

    def push_events_data(self, number, token, data):
        return

    def set_imbalance_and_dsc(self, jobno, rps, comment, ts=0):
        return



'''
Missile object and generators

You should update Stepper.status.ammo_count and Stepper.status.loop_count in your custom generators!
'''
import logging
from itertools import cycle

from yandextank.contrib.netort.netort.resource import manager as resource

from . import info
from .info import LoopCountLimit
from .module_exceptions import AmmoFileError


class HttpAmmo(object):
    '''
    Represents HTTP missile

    >>> print HttpAmmo('/', []).to_s()  # doctest: +NORMALIZE_WHITESPACE
    GET / HTTP/1.1

    >>> print HttpAmmo('/', ['Connection: Close', 'Content-Type: Application/JSON']).to_s()  # doctest: +NORMALIZE_WHITESPACE
    GET / HTTP/1.1
    Connection: Close
    Content-Type: Application/JSON

    >>> print HttpAmmo('/', ['Connection: Close'], method='POST', body='hello!').to_s()  # doctest: +NORMALIZE_WHITESPACE
    POST / HTTP/1.1
    Connection: Close
    Content-Length: 6
    <BLANKLINE>
    hello!
    '''

    def __init__(self, uri, headers, method='GET', http_ver='1.1', body=b''):
        self.method = method
        self.uri = uri.encode('utf8') if isinstance(uri, str) else uri
        self.proto = 'HTTP/%s' % http_ver
        self.headers = set(headers)
        self.body = body
        if len(body):
            self.headers.add("Content-Length: %s" % len(body))

    def to_s(self):
        if self.headers:
            headers = b'\r\n'.join(sorted([h.encode('utf8') for h in self.headers])) + b'\r\n'
        else:
            headers = b''
        return b"%s %s %s\r\n%s\r\n%s" % (
            self.method.encode('utf8'),
            self.uri,
            self.proto.encode('utf8'),
            headers,
            self.body)


class UriStyleGenerator(object):
    """
    Generates GET ammo based on given URI list.
    """

    def __init__(self, uris, headers, http_ver='1.1'):
        """
        uris - a list of URIs as strings.
        """
        super().__init__()
        self.missiles = cycle(
            [(HttpAmmo(uri, headers, http_ver=http_ver).to_s(), None) for uri in uris] + [None]  # None marks loop end
        )

    def __iter__(self):
        for m in self.missiles:
            if m is not None:
                yield m
            else:
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


class Reader(object):
    def __init__(self, filename, use_cache=True, resource_manager=None, **kwargs):
        self.filename = filename
        self.use_cache = use_cache
        self.resource_manager = resource_manager or resource


class AmmoFileReader(Reader):
    """Read missiles from ammo file"""

    def __init__(self, filename, use_cache=True, resource_manager=None, **kwargs):
        super(AmmoFileReader, self).__init__(filename, use_cache, resource_manager)
        self.log = logging.getLogger(__name__)
        self.log.info("Loading ammo from '%s'" % filename)

    @staticmethod
    def read_chunk_header(ammo_file):
        chunk_header = b''
        while chunk_header == b'':
            line = ammo_file.readline()
            if line == b'':
                return line
            chunk_header = line.strip(b'\r\n')
        return chunk_header

    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            chunk_header = self.read_chunk_header(ammo_file)
            while chunk_header:
                if chunk_header != b'':
                    try:
                        fields = chunk_header.split()
                        chunk_size = int(fields[0])
                        if chunk_size == 0:
                            if info.status.loop_count == 0:
                                self.log.info(
                                    'Zero-sized chunk in ammo file at %s. Starting over.'
                                    % ammo_file.tell())
                            ammo_file.seek(0)
                            try:
                                info.status.inc_loop_count()
                            except LoopCountLimit:
                                break
                            chunk_header = self.read_chunk_header(ammo_file)
                            continue
                        marker = fields[1] if len(fields) > 1 else None
                        missile = ammo_file.read(chunk_size)
                        if len(missile) < chunk_size:
                            raise AmmoFileError(
                                "Unexpected end of file: read %s bytes instead of %s"
                                % (len(missile), chunk_size))
                        yield (missile, marker)
                    except (IndexError, ValueError) as e:
                        raise AmmoFileError(
                            "Error while reading ammo file. Position: %s, header: '%s', original exception: %s"
                            % (ammo_file.tell(), chunk_header, e))
                chunk_header = self.read_chunk_header(ammo_file)
                if chunk_header == b'':
                    ammo_file.seek(0)
                    try:
                        info.status.inc_loop_count()
                    except LoopCountLimit:
                        break
                    chunk_header = self.read_chunk_header(ammo_file)
                info.status.af_position = ammo_file.tell()


class SlowLogReader(Reader):
    """Read missiles from SQL slow log. Not usable with Phantom"""
    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            request = ""
            while True:
                for line in ammo_file:
                    info.status.af_position = ammo_file.tell()
                    if isinstance(line, bytes):
                        line = line.decode('utf-8')
                    if line.startswith('#'):
                        if request != "":
                            yield (request, None)
                            request = ""
                    else:
                        request += line
                ammo_file.seek(0)
                info.status.af_position = 0
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


class LineReader(Reader):
    """One line -- one missile"""

    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            while True:
                for line in ammo_file:
                    info.status.af_position = ammo_file.tell()
                    yield (line.rstrip(b'\r\n'), None) if isinstance(line, bytes) else (line.rstrip('\r\n').encode('utf8'), None)
                ammo_file.seek(0)
                info.status.af_position = 0
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


class CaseLineReader(Reader):
    """One line -- one missile with case, tab separated"""

    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            while True:
                for line in ammo_file:
                    info.status.af_position = ammo_file.tell()
                    parts = line.rstrip(b'\r\n').split(b'\t', 1)
                    if len(parts) == 2:
                        yield (parts[1], parts[0])
                    elif len(parts) == 1:
                        yield (parts[0], None)
                    else:
                        raise RuntimeError("Unreachable branch")
                ammo_file.seek(0)
                info.status.af_position = 0
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


class AccessLogReader(Reader):
    """Missiles from access log"""

    def __init__(self, filename, headers=None, http_ver='1.1', use_cache=True, resource_manager=None, **kwargs):
        super(AccessLogReader, self).__init__(filename, use_cache, resource_manager)
        self.warned = False
        self.headers = set(headers) if headers else set()
        self.log = logging.getLogger(__name__)

    def warn(self, message):
        if not self.warned:
            self.warned = True
            self.log.warning("There are some skipped lines. See full log for details.")
        self.log.debug(message)

    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            while True:
                for line in ammo_file:
                    info.status.af_position = ammo_file.tell()
                    if isinstance(line, bytes):
                        line = line.decode('utf-8')
                    try:
                        request = line.split('"')[1]
                        method, uri, proto = request.split()
                        http_ver = proto.split('/')[1]
                        if method == "GET":
                            yield (
                                HttpAmmo(
                                    uri,
                                    headers=self.headers,
                                    http_ver=http_ver, ).to_s(), None)
                        else:
                            self.warn("Skipped line: %s (unsupported method)" % line)
                    except (ValueError, IndexError) as e:
                        self.warn("Skipped line: %s (%s)" % (line, e))
                ammo_file.seek(0)
                info.status.af_position = 0
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


def _parse_header(header):
    return dict([(h.strip().decode('utf8') for h in header.split(b':', 1))])


class UriReader(Reader):
    def __init__(self, filename, headers=None, http_ver='1.1', use_cache=True, resource_manager=None, **kwargs):
        super(UriReader, self).__init__(filename, use_cache, resource_manager)
        self.headers = {pair[0].strip(): pair[1].strip() for pair in [h.split(':', 1) for h in headers]} \
            if headers else {}
        self.http_ver = http_ver
        self.log = logging.getLogger(__name__)
        self.log.info("Loading ammo from '%s' using URI format." % filename)

    def __iter__(self):
        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            while True:
                for line in ammo_file:
                    info.status.af_position = ammo_file.tell()
                    # if isinstance(line, bytes):
                    #     line = line.decode('utf-8')
                    if line.startswith(b'['):
                        self.headers.update(
                            _parse_header(line.strip(b'\r\n[]\t ')))
                    elif len(line.rstrip(b'\r\n')):
                        fields = line.split()
                        uri = fields[0]
                        if len(fields) > 1:
                            marker = fields[1]
                        else:
                            marker = None
                        yield (
                            HttpAmmo(
                                uri,
                                headers=[
                                    ': '.join(header)
                                    for header in self.headers.items()
                                ],
                                http_ver=self.http_ver, ).to_s(), marker)
                if info.status.ammo_count == 0:
                    self.log.error("No ammo in uri-style file")
                    raise AmmoFileError("No ammo! Cover me!")
                ammo_file.seek(0)
                info.status.af_position = 0
                try:
                    info.status.inc_loop_count()
                except LoopCountLimit:
                    break


class UriPostReader(Reader):
    """Read POST missiles from ammo file"""

    def __init__(self, filename, headers=None, http_ver='1.1', use_cache=True, **kwargs):
        super(UriPostReader, self).__init__(filename, use_cache)
        self.headers = {pair[0].strip(): pair[1].strip() for pair in [h.split(':', 1) for h in headers]} \
            if headers else {}
        self.http_ver = http_ver
        self.log = logging.getLogger(__name__)
        self.log.info("Loading ammo from '%s' using URI+POST format", filename)

    def __iter__(self):
        def read_chunk_header(ammo_file):
            chunk_header = b''
            while chunk_header == b'':
                line = ammo_file.readline()
                if line.startswith(b'['):
                    self.headers.update(_parse_header(line.strip(b'\r\n[]\t ')))
                elif line == b'':
                    return line
                else:
                    chunk_header = line.strip(b'\r\n')
            return chunk_header

        opener = self.resource_manager.get_opener(self.filename)
        with opener(self.use_cache) as ammo_file:
            info.status.af_size = opener.data_length
            # if we got StopIteration here, the file is empty
            chunk_header = read_chunk_header(ammo_file)
            while chunk_header:
                if chunk_header != b'':
                    try:
                        fields = chunk_header.split()
                        chunk_size = int(fields[0])
                        uri = fields[1]
                        marker = fields[2] if len(fields) > 2 else None
                        if chunk_size == 0:
                            missile = b""
                        else:
                            missile = ammo_file.read(chunk_size)
                        if len(missile) < chunk_size:
                            raise AmmoFileError(
                                "Unexpected end of file: read %s bytes instead of %s"
                                % (len(missile), chunk_size))
                        yield (
                            HttpAmmo(
                                uri=uri,
                                headers=[
                                    ': '.join(header)
                                    for header in self.headers.items()
                                ],
                                method='POST',
                                body=missile,
                                http_ver=self.http_ver, ).to_s(), marker)
                    except (IndexError, ValueError) as e:
                        raise AmmoFileError(
                            "Error while reading ammo file. Position: %s, header: '%s', original exception: %s"
                            % (ammo_file.tell(), chunk_header, e))
                chunk_header = read_chunk_header(ammo_file)
                if chunk_header == b'':
                    self.log.debug(
                        'Reached the end of ammo file. Starting over.')
                    ammo_file.seek(0)
                    try:
                        info.status.inc_loop_count()
                    except LoopCountLimit:
                        break
                    chunk_header = read_chunk_header(ammo_file)
                info.status.af_position = ammo_file.tell()


FILE_READERS = {
    'phantom': AmmoFileReader,
    'slowlog': SlowLogReader,
    'line': LineReader,
    'uri': UriReader,
    'uripost': UriPostReader,
    'access': AccessLogReader,
    'caseline': CaseLineReader,
}



'''
Contains shellexec plugin
'''
from yandextank.contrib.netort.netort.process import execute
from ...common.interfaces import AbstractPlugin


class Plugin(AbstractPlugin):
    """
    ShellExec plugin
    allows executing shell scripts before/after test
    """
    SECTION = 'shellexec'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.catch_out = False
        self.end = None
        self.poll = None
        self.prepare = None
        self.start = None
        self.postprocess = None
        self.executable = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return ['prepare', 'start', 'end', 'poll', 'post_process', 'catch_out', 'shell']

    def configure(self):
        self.catch_out = True if self.get_option("catch_out", False) else False
        self.prepare = self.get_option("prepare", '')
        self.start = self.get_option("start", '')
        self.end = self.get_option("end", '')
        self.poll = self.get_option("poll", '')
        self.postprocess = self.get_option("post_process", '')
        self.executable = self.get_option('shell', '')

    def prepare_test(self):
        if self.prepare:
            self.execute(self.prepare)

    def start_test(self):
        if self.start:
            self.execute(self.start)

    def is_test_finished(self):
        if self.poll:
            self.log.info("Executing: %s", self.poll)
            retcode = execute(
                self.poll,
                shell=True,
                poll_period=0.1,
                catch_out=self.catch_out,
                executable=self.executable)[0]
            if retcode:
                self.log.warning("Non-zero exit code, interrupting test: %s", retcode)
                return retcode
        return -1

    def end_test(self, retcode):
        if self.end:
            self.execute(self.end)
        return retcode

    def post_process(self, retcode):
        if self.postprocess:
            self.execute(self.postprocess)
        return retcode

    def execute(self, cmd):
        """
        Execute and check exit code
        """
        self.log.info("Executing: %s", cmd)
        retcode = execute(
            cmd, shell=True, poll_period=0.1, catch_out=self.catch_out, executable=self.executable)[0]
        if retcode:
            raise RuntimeError("Subprocess returned %s" % retcode)
        return retcode



"""
Split incoming DataFrames into chunks, cache them, union chunks with same key
and pass to the underlying aggregator.
"""

import pandas as pd


class TimeChopper(object):
    """
    TimeChopper splits incoming dataframes by index. Chunks are cached and
    chunks for same key from different DFs are joined. Then chunks are passed
    further as (<timestamp>, <dataframe>) tuples.
    """

    def __init__(self, sources):
        # self.cache_size = cache_size
        self.sources = {i: src for i, src in enumerate(sources)}
        self.recent_ts = {i: 0 for i in range(len(self.sources))}
        self.cache = {}

    def __iter__(self):
        for _ in range(len(self.sources)):
            try:
                while True:
                    for n, source in self.sources.items():
                        chunk = next(source)
                        if chunk is not None:
                            self.recent_ts[n] = chunk.index[-1]
                            grouped = chunk.groupby(level=0)
                            for ts, group_data in list(grouped):
                                if ts in self.cache:
                                    self.cache[ts] = pd.concat(
                                        [self.cache[ts], group_data])
                                else:
                                    self.cache[ts] = group_data
                    last_ready_ts = min(self.recent_ts.values()) - 1
                    for ts in sorted(filter(lambda x: x <= last_ready_ts, self.cache)):
                        data = self.cache.pop(ts)
                        yield ts, data, len(data)
            except StopIteration:
                self.sources.pop(n)
                self.recent_ts.pop(n)
        while self.cache:
            yield self.__get_result()

    def __get_result(self):
        ts = min(self.cache.keys())
        result = self.cache.pop(ts, None)
        cardinality = len(result) if result is not None else 0
        return ts, result, cardinality



class StepperConfigurationError(Exception):
    '''
    Raised when error in stepper configuration found.
    '''


class AmmoFileError(Exception):
    '''
    Raised when failed to read ammo file properly.
    '''


class StpdFileError(Exception):
    '''
    Raised when failed to read stpd file properly.
    '''


class DiskLimitError(Exception):
    pass



import logging
import time
from threading import Event, Thread

from .guns import LogGun, SqlGun, CustomGun, HttpGun, ScenarioGun, UltimateGun
from .reader import BfgReader, BfgStatsReader
from .widgets import BfgInfoWidget
from ..Phantom import PhantomReader, string_to_df
from .worker import BFGMultiprocessing, BFGGreen
from ..Console import Plugin as ConsolePlugin
from ...common.interfaces import GeneratorPlugin
from ...common.util import FileMultiReader
from ...stepper import StepperWrapper


class Plugin(GeneratorPlugin):
    """ Big Fucking Gun plugin """
    SECTION = 'bfg'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.close_event = Event()
        self._bfg = None
        self.log = logging.getLogger(__name__)
        self.gun_type = None
        self.stepper_wrapper = StepperWrapper(core, cfg)
        self.log.info("Initialized BFG")
        self.report_filename = "bfgout.log"
        self.results_listener = None

        self.gun_classes = {
            'log': LogGun,
            'sql': SqlGun,
            'custom': CustomGun,
            'http': HttpGun,
            'scenario': ScenarioGun,
            'ultimate': UltimateGun,
        }

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "gun_type", "instances", "cached_stpd", "pip"
        ]

    def configure(self):
        self.log.info("Configuring BFG...")
        self.stepper_wrapper.read_config()
        self.stepper_wrapper.prepare_stepper()
        with open(self.report_filename, 'w'):
            pass
        self.core.add_artifact_file(self.report_filename)

    def _write_results_into_file(self):
        """listens for messages on the q, writes to file. """
        reader = BfgReader(self.bfg.results, self.close_event)
        columns = ['receive_ts', 'tag', 'interval_real', 'connect_time', 'send_time', 'latency', 'receive_time',
                   'interval_event', 'size_out', 'size_in', 'net_code', 'proto_code']
        for entry in reader:
            if entry is not None:
                entry.receive_ts = entry.receive_ts.round(3)
                with open(self.report_filename, 'a') as report_file:
                    report_file.write(entry.to_csv(index=False, header=False, sep='\t', columns=columns))
            time.sleep(0.1)

    def get_reader(self, parser=string_to_df):
        if self.reader is None:
            self.reader = FileMultiReader(self.report_filename, self.close_event)
        return PhantomReader(self.reader.get_file(), parser=parser)

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = BfgStatsReader(self.bfg.instance_counter, self.stepper_wrapper.steps)
        return self.stats_reader

    @property
    def bfg(self):
        if self._bfg is None:
            BFG = BFGGreen if self.get_option("worker_type", "") == "green" else BFGMultiprocessing
            self._bfg = BFG(
                gun=self.gun,
                instances=self.stepper_wrapper.instances,
                stpd_filename=self.stepper_wrapper.stpd,
                cached_stpd=self.get_option("cached_stpd"),
                green_threads_per_instance=int(self.get_option('green_threads_per_instance', 1000)),
            )
        return self._bfg

    def prepare_test(self):
        pip_deps = self.get_option("pip", "").splitlines()
        self.log.info("Installing with PIP: %s", pip_deps)
        if pip_deps:
            import pip
            retcode = pip.main(["install", "--user"] + pip_deps)
            if retcode != 0:
                raise RuntimeError("Could not install required deps")
            import site
            from importlib import reload
            reload(site)
        self.log.info("BFG using ammo type %s", self.get_option("ammo_type"))
        gun_type = self.get_option("gun_type")
        if gun_type in self.gun_classes:
            self.gun = self.gun_classes[gun_type](self.core, self.get_option('gun_config'))
        else:
            raise NotImplementedError('No such gun type implemented: "%s"' % gun_type)

        self.results_listener = Thread(target=self._write_results_into_file, name="ResultsQueueListener")

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            self.log.debug(ex)
            console = None

        if console:
            widget = BfgInfoWidget()
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)
        self.log.info("Prepared BFG")

    def start_test(self):
        self.log.info("Starting BFG")
        self.start_time = time.time()
        self.bfg.start()
        if self.results_listener is not None:
            self.results_listener.start()
        else:
            self.log.fatal("Result listener is not initialized")

    def is_test_finished(self):
        if self.bfg.running():
            return -1
        else:
            self.log.info("BFG finished")
            self.close_event.set()
            self.stats_reader.close()
            return 0

    def end_test(self, retcode):
        if self.bfg.running():
            self.log.info("Terminating BFG")
            self.bfg.stop()
        self.close_event.set()
        self.stats_reader.close()
        return retcode



import pytest
import re
from yandextank.common.interfaces import TankInfo, AbstractCriterion


class TestStatus(object):

    @pytest.mark.parametrize('updates, result', [
        ([(['plugin', 'key1'], 'foo'), (['plugin', 'key2'], 42)], {'plugin': {'key1': 'foo', 'key2': 42}}),
        ([(['plugin1', 'key1'], 'foo'),
          (['plugin1', 'key2'], 42),
          (['plugin2', 'key1'], 'bar')], {'plugin1': {'key1': 'foo', 'key2': 42},
                                          'plugin2': {'key1': 'bar'}})
    ])
    def test_update(self, updates, result):
        info = TankInfo(dict())
        for args in updates:
            info.update(*args)
        assert info.get_info_dict() == result


@pytest.mark.parametrize('codes_mask, codes_dict, expected_matched_cnt',
                         [
                             ('', {'200': 500}, 0),  # mask is empty
                             ('1', {'110': 500}, 0),  # mask is prefix
                             ('1', {'21': 500}, 0),  # mask is suffix
                             ('0.', {'0': 500}, 0),  # mask is too long
                             ('2.', {'200': 500}, 0),  # mask is too short
                             ('2..', {'200': 500}, 500),
                             ('2..', {'200': 500, '201': 100}, 600),
                             ('..9', {'999': 500}, 500),
                         ])
def test_match(codes_mask, codes_dict, expected_matched_cnt):
    assert AbstractCriterion.count_matched_codes(codes_regex=re.compile(codes_mask),
                                                 codes_dict=codes_dict) == expected_matched_cnt



import logging
import typing


class AbstractPlugin(object):
    """ Plugin interface
    Parent class for all plugins """

    SECTION = 'DEFAULT'

    @staticmethod
    def get_key():
        """ Get dictionary key for plugin,
        should point to __file__ magic constant """
        raise TypeError("Abstract method needs to be overridden")

    # TODO: do we realy need cfg_updater here?
    def __init__(self, core, cfg, name):
        """

        :param name:
        :type core: yandextank.core.TankCore
        :type cfg: dict
        """
        super(AbstractPlugin, self).__init__()
        self._cleanup_actions = []
        self.log = logging.getLogger(__name__)
        self.core = core
        self.cfg = cfg
        self.cfg_section_name = name
        self.interrupted = self.core.interrupted
        self.errors = []
        self.resource_manager = self.core.resource_manager

    def set_option(self, option, value):
        self.cfg[option] = value

    def configure(self):
        """ A stage to read config values and instantiate objects """
        pass

    def prepare_test(self):
        """        Test preparation tasks        """
        pass

    def start_test(self):
        """        Launch test process        """
        pass

    def is_test_finished(self):
        """
        Polling call, if result differs from -1 then test end
        will be triggered
        """
        return -1

    def add_cleanup(self, action):
        """
        :type action: function
        """
        assert callable(action)
        self._cleanup_actions.append(action)

    def cleanup(self):
        for action in reversed(self._cleanup_actions):
            try:
                action()
            except Exception:
                logging.exception('Exception occurred during %s plugin cleanup', self.__module__)

    def end_test(self, retcode):
        """
        Stop processes launched at 'start_test',
        change return code if necessary
        """
        return retcode

    def post_process(self, retcode):
        """ Post-process test data """
        return retcode

    def get_option(self, option_name, default_value=None):
        """ Wrapper to get option from plugins' section """
        return self.cfg.get(option_name, default_value)

    def get_available_options(self):
        """ returns array containing known options for plugin """
        return []

    def get_multiline_option(self, option_name, default_value=None):
        if default_value is not None:
            default = ' '.join(default_value)
        else:
            default = None
        value = self.get_option(option_name, default)
        if value:
            return (' '.join(value.split("\n"))).split(' ')
        else:
            return ()

    def publish(self, key, value):
        """publish value to status"""
        self.log.debug(
            "Publishing status: %s/%s: %s", self.__class__.SECTION, key, value)
        self.core.publish(self.__class__.SECTION, key, value)

    def close(self):
        """
        Release allocated resources here.
        Warning: don't do any logic or potentially dangerous operations
        """
        pass


class MonitoringDataListener(object):
    """ Monitoring listener interface
    parent class for Monitoring data listeners"""

    def __init__(self):
        pass

    def monitoring_data(self, data):
        """Notification about new monitoring data lines"""
        raise NotImplementedError("Abstract method needs to be overridden")


class AggregateResultListener(object):
    """ Listener interface
    parent class for Aggregate results listeners"""

    def on_aggregated_data(self, data, stats):
        """
        notification about new aggregated data and stats

        data contains aggregated metrics and stats contain non-aggregated
        metrics from gun (like instances count, for example)

        data and stats are cached and synchronized by timestamp. Stat items
        are holded until corresponding data item is received and vice versa.
        """
        raise NotImplementedError("Abstract method should be overridden")


class AbstractInfoWidget(object):
    """ InfoWidgets interface
    parent class for all InfoWidgets"""

    def __init__(self):
        pass

    def render(self, screen):
        raise NotImplementedError("Abstract method should be overridden")

    def on_aggregated_data(self, data, stats):
        raise NotImplementedError("Abstract method should be overridden")

    def get_index(self):
        """ get vertical priority index """
        return 0


class AbstractCriterion(object):
    """ Criterions interface,
    parent class for all criterions """
    RC_TIME = 21
    RC_HTTP = 22
    RC_NET = 23
    RC_INSTANCE = 24
    RC_TOTAL_TIME = 25
    RC_TOTAL_HTTP = 26
    RC_TOTAL_NET = 27
    RC_TOTAL_NEGATIVE_HTTP = 28
    RC_TOTAL_NEGATIVE_NET = 29
    RC_TOTAL_HTTP_TREND = 30
    RC_METRIC_HIGHER = 31
    RC_METRIC_LOWER = 32
    RC_STEADY = 33

    def __init__(self):
        self.cause_second = None

    @staticmethod
    def count_matched_codes(codes_regex, codes_dict):
        """ helper to aggregate codes by mask """
        total = 0
        for code, count in codes_dict.items():
            if codes_regex.fullmatch(str(code)):
                total += count
        return total

    def notify(self, data, stat):
        """ notification about aggregate data goes here """
        raise NotImplementedError("Abstract methods requires overriding")

    def get_rc(self):
        """ get return code for test """
        raise NotImplementedError("Abstract methods requires overriding")

    def explain(self):
        """ long explanation to show after test stop """
        raise NotImplementedError("Abstract methods requires overriding")

    def get_criterion_parameters(self):
        """ returns dict with all criterion parameters """
        raise NotImplementedError("Abstract methods requires overriding")

    def widget_explain(self):
        """ short explanation to display in right panel """
        return self.explain(), 0

    @staticmethod
    def get_type_string():
        """ returns string that used as config name for criterion """
        raise NotImplementedError("Abstract methods requires overriding")


class GeneratorPlugin(AbstractPlugin):
    DEFAULT_INFO = {
        'address': '',
        'port': 80,
        'instances': 1,
        'ammo_file': '',
        'rps_schedule': [],
        'duration': 0,
        'loop_count': 0
    }

    def __init__(self, core, cfg, name):
        super(GeneratorPlugin, self).__init__(core, cfg, name)
        self.stats_reader = None
        self.reader = None
        self.process = None
        self.process_stderr = None
        self.start_time = None
        self.affinity = None
        self.buffered_seconds = 2

    class Info(object):
        def __init__(
                self, address, port, instances, ammo_file, rps_schedule,
                duration, loop_count):
            self.address = address
            self.port = port
            self.instances = instances
            self.ammo_file = ammo_file
            self.rps_schedule = rps_schedule
            self.duration = duration
            self.loop_count = loop_count

    def get_info(self):
        """
        :rtype: GeneratorPlugin.Info
        """
        return self.Info(**self.DEFAULT_INFO)

    def get_reader(self):
        """

        :rtype: collections.Iterable
        """
        pass

    def get_stats_reader(self):
        """

        :rtype: collections.Iterable
        """
        pass

    def end_test(self, retcode):
        pass


class StatsReader(object):
    @staticmethod
    def stats_item(ts, instances, rps):
        return {
            'ts': ts,
            'metrics': {
                'instances': instances,
                'reqps': rps
            }
        }


class MonitoringPlugin(AbstractPlugin):
    def __init__(self, core, cfg, name):
        super(MonitoringPlugin, self).__init__(core, cfg, name)
        self.listeners = set()
        self.collector = None

    def add_listener(self, plugin):
        self.listeners.add(plugin)

    def start_test(self):   # noqa: PLE0202
        if self.collector is None:
            self.collector: MonitoringCollectorProtocol = DummyCollector()
        self.collector.start()

    def end_test(self, retcode):   # noqa: PLE0202
        self.collector.stop()
        self.poll()
        return retcode

    def is_test_finished(self):   # noqa: PLE0202
        self.poll()
        return -1

    def poll(self):
        data = self.collector.poll()
        if len(data) > 0:
            self.core.publish_monitoring_data(data)


MonitoringChunk = typing.Collection


class MonitoringCollectorProtocol(typing.Protocol):
    def start(self):
        ...

    def stop(self):
        ...

    def poll(self) -> MonitoringChunk:
        ...


class DummyCollector(MonitoringCollectorProtocol):
    def start(self):
        return

    def stop(self):
        return

    def poll(self) -> MonitoringChunk:
        return []


class TankInfo(object):
    def __init__(self, info):
        self._info = info

    def get_info_dict(self):
        return self._info.copy()

    def _set_info(self, new_info_dict):
        raise NotImplementedError

    def update(self, keys, value):
        if len(keys) > 1:
            self._info[keys[0]] = self._update_dict(self.get_value([keys[0]], {}), keys[1:], value)
        else:
            self._info[keys[0]] = value

    def get_value(self, keys, default=None):
        value = self.get_info_dict()
        for key in keys:
            value = value.get(key, {})
        return value or default

    @classmethod
    def _update_dict(cls, status_dict, keys, value):
        if len(keys) > 1:
            cls._update_dict(status_dict.setdefault(keys[0], {}), keys[1:], value)
        else:
            status_dict[keys[0]] = value
        return status_dict



# coding=utf-8
# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import datetime
import logging
import sys
from builtins import str
from uuid import uuid4

from .client import OpenTSDBClient
from .decoder import Decoder
from ...common.interfaces import AbstractPlugin, \
    MonitoringDataListener, AggregateResultListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def chop(data_list, chunk_size):
    if sys.getsizeof(str(data_list)) <= chunk_size:
        return [data_list]
    elif len(data_list) == 1:
        logger.warning("Too large piece of Telegraf data. Might experience upload problems.")
        return [data_list]
    else:
        mid = len(data_list) / 2
        return chop(data_list[:mid], chunk_size) + chop(
            data_list[mid:], chunk_size)


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    SECTION = 'opentsdb'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.tank_tag = self.get_option("tank_tag")
        self.prefix_metric = self.get_option("prefix_metric")
        self._client = None
        self.start_time = None
        self.end_time = None

        self.decoder = Decoder(
            self.tank_tag,
            str(uuid4()),
            self.get_option("custom_tags"),
            self.get_option("labeled"),
            self.get_option("histograms"),
        )

    @property
    def client(self):
        if not self._client:
            self._client = OpenTSDBClient(
                host=self.get_option("address"),
                port=self.get_option("port"),
                username=self.get_option("username"),
                password=self.get_option("password"),
                ssl=self.get_option("ssl"),
                verify_ssl=self.get_option("verify_ssl"))
        return self._client

    def prepare_test(self):
        self.core.job.subscribe_plugin(self)

    def start_test(self):
        self.start_time = datetime.datetime.now()

    def end_test(self, retcode):
        self.end_time = datetime.datetime.now() + datetime.timedelta(minutes=1)
        return retcode

    def on_aggregated_data(self, data, stats):
        self.client.write(
            self.decoder.decode_aggregates(data, stats, self.prefix_metric))

    def monitoring_data(self, data_list):
        if len(data_list) > 0:
            [
                self._send_monitoring(chunk)
                for chunk in chop(data_list, self.get_option("chunk_size"))
            ]

    def _send_monitoring(self, data):
        self.client.write(self.decoder.decode_monitoring(data))

    def set_uuid(self, id_):
        self.decoder.tags['uuid'] = id_



import logging
import uuid
import time
import os
import getpass
import six
from yandextank.contrib.netort.netort.data_manager.common.util import thread_safe_property
from typing import Callable, Dict, Optional, Set, Any

from .metrics import Metric, Event
from .clients import available_clients
from .router import MetricsRouter
from .common.interfaces import AbstractMetric

if six.PY3:
    from queue import Queue
else:  # six.PY2
    # noinspection PyUnresolvedReferences
    from Queue import Queue

logger = logging.getLogger(__name__)


# TODO: move code that works with config to library's clients (e.g., Volta).
# Classes of this library should provide constructors with described arguments only
# noinspection PyBroadException
class DataSession(object):
    """
    Workflow:
        * create DataSession object
        * use `new_metric` to add metrics to your datasession
        * use `metric.put` to add data to the metric
        * call `close` to close your datasession

    Note:
        * send your data in chunks because it could be of bigger size that server's buffer

    Args:
        config(dict): configuration options (list of DataManager clients, test meta data etc)

    TODO:
        * move config parameters to kwargs, describe them here
        * fight performance issues (probably caused by poor pandas write_csv performance)
    """
    def __init__(self, config: Dict, tankapi_info: Dict = None, config_filenames: Set = None, artifacts_dir: str = None, test_start: float = None):
        self.start_ts = time.time()
        self.config = config
        self.operator = self.__get_operator()
        self.job_id = config.get('test_id', 'job_{uuid}'.format(uuid=uuid.uuid4()))
        logger.info('Created new local data session: %s', self.job_id)
        self.tankapi_info = tankapi_info
        self.config_filenames = config_filenames or set()
        self.test_start = test_start if test_start else int(time.time() * 10**6)
        self.artifacts_base_dir = artifacts_dir or './logs'
        self._artifacts_dir: Optional[str] = None
        self.manager = DataManager()

        self.clients = []
        self.__create_clients(config.get('clients', []))
        logger.debug('DataSession clients: %s', self.clients)

    # TODO: extract client creation as factory method
    # TODO: consider removing clients from config and add them via `new_client` method
    def __create_clients(self, clients: Dict):
        for client_meta in clients:
            type_ = client_meta.get('type')
            if not type_:
                raise ValueError('Client type should be defined.')
            if type_ in available_clients:
                client = available_clients[type_](client_meta, self)
                self.subscribe(client.put)
                self.clients.append(client)
            else:
                raise NotImplementedError('Unknown client type: %s' % type_)

    def new_true_metric(self, meta: Dict, raw=True, aggregate=False, parent: str = None, case: str = None) -> AbstractMetric:
        return self.manager.new_true_metric(meta=meta,
                                            test_start=self.test_start,
                                            raw=raw, aggregate=aggregate,
                                            parent=parent, case=case)

    def new_event_metric(self, meta: Dict, raw=True, aggregate=False, parent: str = None, case: str = None) -> AbstractMetric:
        return self.manager.new_event_metric(meta=meta,
                                             test_start=self.test_start,
                                             raw=raw, aggregate=aggregate,
                                             parent=parent, case=case)

    def subscribe(self, callback: Callable) -> Any:
        return self.manager.subscribe(callback)

    def get_metric_by_id(self, id_: str) -> Any:
        return self.manager.get_metric_by_id(id_)

    def update_job(self, meta: Dict) -> Any:
        for client in self.clients:
            try:
                client.update_job(meta)
            except Exception:
                logger.warning('Client %s job update failed', client)
                logger.debug('Client %s job update failed', client, exc_info=True)
            else:
                logger.debug('Client job updated: %s', client)

    def update_metric(self, meta: Dict) -> None:
        for client in self.clients:
            try:
                client.update_metric(meta)
            except Exception:
                logger.warning('Client %s metric update failed', client)
                logger.debug('Client %s metric update failed', client, exc_info=True)
            else:
                logger.debug('Client metric updated: %s', client)

    # TODO: artifacts dir should be inside "local" client. Or does it?
    @thread_safe_property
    def artifacts_dir(self) -> str:
        if not self._artifacts_dir:
            dir_name = "{dir}/{id}".format(dir=self.artifacts_base_dir, id=self.job_id)
            if not os.path.isdir(dir_name):
                os.makedirs(dir_name)
            os.chmod(dir_name, 0o755)
            self._artifacts_dir = os.path.abspath(dir_name)
        return self._artifacts_dir

    def __get_operator(self) -> str:
        try:
            return self.config.get('operator') or getpass.getuser()
        except:  # noqa: E722
            logger.error(
                "Couldn't get username from the OS. Please, set the 'operator' option explicitly in your config "
                "file.")
            raise

    def close(self, test_end: float):
        logger.info('DataSession received close signal.')
        logger.info('Closing DataManager')
        self.manager.close()
        logger.info('Waiting the rest of data from router...')
        self.manager.router.join()
        logger.info('Sending close to DataSession clients...')
        for client in self.clients:
            try:
                client.close(test_end)
            except Exception:
                logger.warning('Client %s failed to close', client, exc_info=True)
            else:
                logger.debug('Client closed: %s', client)
        logger.info('DataSession finished!')
        logger.info('DataSession time: %s', time.time() - self.start_ts)

    def interrupt(self):
        self.manager.interrupt()
        for client in self.clients:
            try:
                client.close()
            except Exception:
                logger.warning('Client %s failed to close', client)
            else:
                logger.debug('Client closed: %s', client)
        logger.info('DataSession finished!')


class DataManager(object):
    """DataManager routes data to subscribers using metrics meta as a filter. When someone calls
    `new_metric`, DataManager will find the subscribers that are interested in this metric (using meta).
    When someone calls `subscribe`, DataManager finds the metrics that this subscriber is interested in.

    MetricsRouter is a facility that DataManager uses for passing incoming data to subscribers.

    Attributes:
        metrics: All registered metrics for DataManager session
        subscribers: All registered subscribers for DataManager session
        callbacks: callbacks for metric ids <-> subscribers' callbacks, used by router
        routing_queue: incoming unrouted metrics data, will be processed by MetricsRouter to subscribers' callbacks
        router (MetricsRouter object): Router thread. Read routing queue, concat incoming messages by metrics.type,
            left join by callback and call callback w/ resulting dataframe
    """
    def __init__(self):
        self.metrics = {}
        self.metrics_meta = {}
        self.subscribers = {}
        self.callbacks = {}
        self.routing_queue = Queue()
        self.router = MetricsRouter(self)
        self.router.start()

    def new_true_metric(self, meta: Dict, test_start: float, raw=True, aggregate=False, parent: str = None, case: str = None) -> AbstractMetric:
        """
        Create and register metric,
        find subscribers for this metric (using meta as filter) and subscribe

        Return:
            metric: one of Metric
        """
        return self._new_metric(Metric, meta, test_start, raw, aggregate, parent=parent, case=case)

    def new_event_metric(self, meta: Dict, test_start: float, raw=True, aggregate=False, parent: str = None, case: str = None) -> AbstractMetric:
        # type: (dict, float, bool, bool, str, str) -> AbstractMetric
        return self._new_metric(Event, meta, test_start, raw, aggregate, parent=parent, case=case)

    def _new_metric(self, dtype, meta: Dict, test_start: float, raw=True, aggregate=False, parent: str = None, case: str = None) -> AbstractMetric:
        metric_obj = dtype(meta=meta,
                           _queue=self.routing_queue,
                           test_start=test_start,
                           raw=raw, aggregate=aggregate,
                           parent=parent, case=case)  # create metric object
        self.metrics_meta = meta  # register metric meta
        self.metrics[metric_obj.local_id] = metric_obj  # register metric object
        for callback in self.callbacks:
            self.callbacks[callback].add(metric_obj.local_id)
        return metric_obj

    def subscribe(self, callback: Callable):
        """
        Create and register metric subscriber,
        subscribe all existing metrics for it

        Args:
            callback (object method): subscriber's callback
        """
        sub_id = "subscriber_{uuid}".format(uuid=uuid.uuid4())
        # register subscriber in manager
        self.subscribers[sub_id] = callback
        self.callbacks[sub_id] = set(iter(self.metrics))

    def get_metric_by_id(self, id_: str) -> Optional[AbstractMetric]:
        return self.metrics.get(id_)

    def close(self):
        self.router.close()

    def interrupt(self):
        self.router.interrupt()
        self.router.join()


# def usage_sample():
#     import time
#     import pandas as pd
#     config = {
#         'clients': [
#             {
#                 'type': 'luna',
#                 'api_address': 'http://hostname.tld',
#                 'user_agent': 'Tank Test',
#             },
#             {
#                 'type': 'local_storage',
#             }
#         ],
#         'test_start': time.time(),
#         'artifacts_base_dir': './logs'
#     }
#     data_session = DataSession(config=config)
#
#     metric_meta = {
#         'type': 'metrics',
#         'name': 'cpu_usage',
#         'hostname': 'localhost',
#         'some_meta_key': 'some_meta_value'
#     }
#
#     metric_obj = data_session.new_true_metric('name', **metric_meta)
#     time.sleep(1)
#     df = pd.DataFrame([[123, 123.123, "trash"]], columns=['ts', 'value', 'trash'])
#     metric_obj.put(df)
#     df2 = pd.DataFrame([[456, 456.456]], columns=['ts', 'value'])
#     metric_obj.put(df2)
#     time.sleep(10)
#     df = pd.DataFrame([[123, 123.123]], columns=['ts', 'value'])
#     metric_obj.put(df)
#     df2 = pd.DataFrame([[456, 456.456]], columns=['ts', 'value'])
#     metric_obj.put(df2)
#     time.sleep(10)
#     data_session.close(test_end=time.time())


if __name__ == '__main__':
    logging.basicConfig(level='DEBUG')
    logger = logging.getLogger(__name__)
    # usage_sample()



# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import json
import logging
import numpy as np
import os

import io

from ...common.interfaces import AbstractPlugin, \
    MonitoringDataListener, AggregateResultListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    # pylint:disable=R0902
    SECTION = 'json_report'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.monitoring_stream = io.open(os.path.join(self.core.artifacts_dir,
                                                      self.get_option('monitoring_log')),
                                         mode='wb')
        self.data_and_stats_stream = io.open(os.path.join(self.core.artifacts_dir,
                                                          self.get_option('test_data_log')),
                                             mode='wb')
        self._is_telegraf = None

    def get_available_options(self):
        return ['monitoring_log', 'test_data_log']

    def configure(self):
        self.core.job.subscribe_plugin(self)

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """
        json_string = json.dumps({
            'data': data,
            'stats': stats
        }, cls=NumpyEncoder)
        self.data_and_stats_stream.write('{}\n'.format(json_string).encode('utf-8'))

    def monitoring_data(self, data_list):
        if self.is_telegraf:
            monitoring_data = '{}\n'.format(json.dumps(data_list)).encode('utf-8')
            self.monitoring_stream.write(monitoring_data)
        else:
            [
                self.monitoring_stream.write('{}\n'.format(data.strip()).encode('utf-8')) for data in data_list
                if data
            ]

    def post_process(self, retcode):
        self.close()
        return retcode

    def close(self):
        if self.data_and_stats_stream:
            self.data_and_stats_stream.close()
        if self.monitoring_stream:
            self.monitoring_stream.close()

    @property
    def is_telegraf(self):
        return True



# -*- coding: utf-8 -*-
"""Python client for OpenTSDB."""

import json
import random
import requests
import requests.exceptions
import time


class OpenTSDBClient(object):
    """OpenTSDBClient primary client object to connect OpenTSDB.
    The :class:`~.OpenTSDBClient` object holds information necessary to
    connect to OpenTSDB. Requests can be made to OpenTSDB directly through
    the client.
    :param host: hostname to connect to OpenTSDB, defaults to 'localhost'
    :type host: str
    :param port: port to connect to OpenTSDB, defaults to 4242
    :type port: int
    :param username: user to connect, defaults to 'root'
    :type username: str
    :param password: password of the user, defaults to 'root'
    :type password: str
    :param pool_size: urllib3 connection pool size, defaults to 10.
    :type pool_size: int
    :param ssl: use https instead of http to connect to OpenTSDB, defaults to
        False
    :type ssl: bool
    :param verify_ssl: verify SSL certificates for HTTPS requests, defaults to
        False
    :type verify_ssl: bool
    :param timeout: number of seconds Requests will wait for your client to
        establish a connection, defaults to None
    :type timeout: int
    :param retries: number of retries your client will try before aborting,
        defaults to 3. 0 indicates try until success
    :type retries: int
    :param proxies: HTTP(S) proxy to use for Requests, defaults to {}
    :type proxies: dict
    :param cert: Path to client certificate information to use for mutual TLS
        authentication. You can specify a local cert to use
        as a single file containing the private key and the certificate, or as
        a tuple of both files’ paths, defaults to None
    :type cert: str
    :raises ValueError: if cert is provided but ssl is disabled (set to False)
    """
    def __init__(
            self,
            host='localhost',
            port=4242,
            username='root',
            password='root',
            ssl=False,
            verify_ssl=False,
            timeout=None,
            retries=3,
            proxies=None,
            pool_size=10,
            cert=None,
    ):
        """Construct a new OpenTSDBClient object."""
        self._host = host
        self._port = int(port)
        self._username = username
        self._password = password
        self._timeout = timeout
        self._retries = retries

        self._verify_ssl = verify_ssl

        self._session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            pool_connections=int(pool_size), pool_maxsize=int(pool_size))

        self._scheme = "http"

        if ssl is True:
            self._scheme = "https"

        self._session.mount(self._scheme + '://', adapter)

        if proxies is None:
            self._proxies = {}
        else:
            self._proxies = proxies

        if cert:
            if not ssl:
                raise ValueError("Client certificate provided but ssl is disabled.")
            else:
                self._session.cert = cert

        self._baseurl = "{0}://{1}:{2}".format(
            self._scheme, self._host, self._port)

        self._headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

    def write(self, data, expected_response_code=204):
        """Write data to OpenTSDB.
        :param data: the data to be written
        :param expected_response_code: the expected response code of the write
            operation, defaults to 204
        :type expected_response_code: int
        :returns: True, if the write operation is successful
        :rtype: bool
        """
        headers = self._headers
        headers['Content-Type'] = 'application/json'

        self.request(
            url="api/put",
            method='POST',
            data=data,
            expected_response_code=expected_response_code,
            headers=headers)
        return True

    def request(
            self,
            url,
            method='GET',
            params=None,
            data=None,
            expected_response_code=200,
            headers=None):
        """Make a HTTP request to the OpenTSDB API.
        :param url: the path of the HTTP request
        :type url: str
        :param method: the HTTP method for the request, defaults to GET
        :type method: str
        :param params: additional parameters for the request, defaults to None
        :type params: dict
        :param data: the data of the request, defaults to None
        :type data: str
        :param expected_response_code: the expected response code of
            the request, defaults to 200
        :type expected_response_code: int
        :param headers: headers to add to the request
        :type headers: dict
        :returns: the response from the request
        :rtype: :class:`requests.Response`
        :raises OpenTSDBServerError: if the response code is any server error
            code (5xx)
        :raises OpenTSDBClientError: if the response code is not the
            same as `expected_response_code` and is not a server error code
        """
        url = "{0}/{1}".format(self._baseurl, url)

        if headers is None:
            headers = self._headers

        if params is None:
            params = {}

        if isinstance(data, (dict, list)):
            data = json.dumps(data)

        # Try to send the request more than once by default (see #103)
        retry = True
        _try = 0
        while retry:
            try:
                response = self._session.request(
                    method=method,
                    url=url,
                    auth=(self._username, self._password),
                    params=params,
                    data=data,
                    headers=headers,
                    proxies=self._proxies,
                    verify=self._verify_ssl,
                    timeout=self._timeout)
                break
            except (requests.exceptions.ConnectionError,
                    requests.exceptions.HTTPError, requests.exceptions.Timeout):
                _try += 1
                if self._retries != 0:
                    retry = _try < self._retries
                if method == "POST":
                    time.sleep((2**_try) * random.random() / 100.0)
                if not retry:
                    raise
        # if there's not an error, there must have been a successful response
        if 500 <= response.status_code < 600:
            raise Exception(response.content)
        elif response.status_code == expected_response_code:
            return response
        else:
            raise Exception(response.content, response.status_code)
